import io
import json
import os
import pathlib

from typing import List, Optional, TypeVar, Union, cast
from typing_extensions import ParamSpec

from weaviate.collections.classes.aggregate import (
    AProperties,
    AggregateResult,
    AggregateBoolean,
    AggregateDate,
    AggregateInteger,
    AggregateNumber,
    # AggregateReference, # Aggregate references currently bugged on Weaviate's side
    AggregateText,
    AggregateGroup,
    AggregateGroupByReturn,
    AggregateReturn,
    GroupByAggregate,
    _Metrics,
    _MetricsBoolean,
    _MetricsDate,
    _MetricsNumber,
    _MetricsInteger,
    # _MetricsReference, # Aggregate references currently bugged on Weaviate's side
    _MetricsText,
    GroupedBy,
    TopOccurrence,
)
from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.collections.classes.filters import _Filters
from weaviate.collections.classes.grpc import Move
from weaviate.connect import ConnectionV4
from weaviate.collections.filters import _FilterToREST
from weaviate.exceptions import WeaviateInvalidInputError, WeaviateQueryError
from weaviate.gql.aggregate import AggregateBuilder
from weaviate.util import file_encoder_b64
from weaviate.validator import _ValidateArgument, _validate_input
from weaviate.types import NUMBER, UUID

P = ParamSpec("P")
T = TypeVar("T")


class _Aggregate:
    def __init__(
        self,
        connection: ConnectionV4,
        name: str,
        consistency_level: Optional[ConsistencyLevel],
        tenant: Optional[str],
    ):
        self._connection = connection
        self.__name = name
        self._tenant = tenant
        self._consistency_level = consistency_level

    def _query(self) -> AggregateBuilder:
        return AggregateBuilder(self.__name, self._connection)

    def _to_aggregate_result(
        self, response: dict, metrics: Optional[List[_Metrics]]
    ) -> AggregateReturn:
        try:
            result: dict = response["data"]["Aggregate"][self.__name][0]
            return AggregateReturn(
                properties=self.__parse_properties(result, metrics) if metrics is not None else {},
                total_count=result["meta"]["count"] if result.get("meta") is not None else None,
            )
        except KeyError as e:
            raise ValueError(
                f"There was an error accessing the {e} key when parsing the GraphQL response: {response}"
            )

    def _to_group_by_result(
        self, response: dict, metrics: Optional[List[_Metrics]]
    ) -> AggregateGroupByReturn:
        try:
            results: dict = response["data"]["Aggregate"][self.__name]
            return AggregateGroupByReturn(
                groups=[
                    AggregateGroup(
                        grouped_by=GroupedBy(
                            prop=result["groupedBy"]["path"][0],
                            value=result["groupedBy"]["value"],
                        ),
                        properties=(
                            self.__parse_properties(result, metrics) if metrics is not None else {}
                        ),
                        total_count=(
                            result["meta"]["count"] if result.get("meta") is not None else None
                        ),
                    )
                    for result in results
                ]
            )
        except KeyError as e:
            raise ValueError(
                f"There was an error accessing the {e} key when parsing the GraphQL response: {response}"
            )

    def __parse_properties(self, result: dict, metrics: List[_Metrics]) -> AProperties:
        props: AProperties = {}
        for metric in metrics:
            if metric.property_name in result:
                props[metric.property_name] = self.__parse_property(
                    result[metric.property_name], metric
                )
        return props

    @staticmethod
    def __parse_property(property_: dict, metric: _Metrics) -> AggregateResult:
        if isinstance(metric, _MetricsText):
            return AggregateText(
                count=property_.get("count"),
                top_occurrences=[
                    TopOccurrence(
                        count=cast(dict, top_occurrence).get("occurs"),
                        value=cast(dict, top_occurrence).get("value"),
                    )
                    for top_occurrence in property_.get("topOccurrences", [])
                ],
            )
        elif isinstance(metric, _MetricsInteger):
            return AggregateInteger(
                count=property_.get("count"),
                maximum=property_.get("maximum"),
                mean=property_.get("mean"),
                median=property_.get("median"),
                minimum=property_.get("minimum"),
                mode=property_.get("mode"),
                sum_=property_.get("sum"),
            )
        elif isinstance(metric, _MetricsNumber):
            return AggregateNumber(
                count=property_.get("count"),
                maximum=property_.get("maximum"),
                mean=property_.get("mean"),
                median=property_.get("median"),
                minimum=property_.get("minimum"),
                mode=property_.get("mode"),
                sum_=property_.get("sum"),
            )
        elif isinstance(metric, _MetricsBoolean):
            return AggregateBoolean(
                count=property_.get("count"),
                percentage_false=property_.get("percentageFalse"),
                percentage_true=property_.get("percentageTrue"),
                total_false=property_.get("totalFalse"),
                total_true=property_.get("totalTrue"),
            )
        elif isinstance(metric, _MetricsDate):
            return AggregateDate(
                count=property_.get("count"),
                maximum=property_.get("maximum"),
                median=property_.get("median"),
                minimum=property_.get("minimum"),
                mode=property_.get("mode"),
            )
        # Aggregate references currently bugged on Weaviate's side
        # elif isinstance(metric, _MetricsReference):
        #     return AggregateReference(pointing_to=property_.get("pointingTo"))
        else:
            raise ValueError(
                f"Unknown aggregation type {metric} encountered in _Aggregate.__parse_property() for property {property_}"
            )

    @staticmethod
    def _add_groupby_to_builder(
        builder: AggregateBuilder, group_by: Union[str, GroupByAggregate, None]
    ) -> AggregateBuilder:
        _validate_input(_ValidateArgument([str, GroupByAggregate, None], "group_by", group_by))
        if group_by is None:
            return builder
        if isinstance(group_by, str):
            group_by = GroupByAggregate(prop=group_by)
        builder = builder.with_group_by_filter([group_by.prop])
        if group_by.limit is not None:
            builder = builder.with_limit(group_by.limit)
        return builder.with_fields(" groupedBy { path value } ")

    def _base(
        self,
        return_metrics: Optional[List[_Metrics]],
        filters: Optional[_Filters],
        total_count: bool,
    ) -> AggregateBuilder:
        _validate_input(
            [
                _ValidateArgument([List[_Metrics], None], "return_metrics", return_metrics),
                _ValidateArgument([_Filters, None], "filters", filters),
                _ValidateArgument([bool], "total_count", total_count),
            ]
        )
        builder = self._query()
        if return_metrics is not None:
            builder = builder.with_fields(" ".join([metric.to_gql() for metric in return_metrics]))
        if filters is not None:
            builder = builder.with_where(_FilterToREST.convert(filters))
        if total_count:
            builder = builder.with_meta_count()
        if self._tenant is not None:
            builder = builder.with_tenant(self._tenant)
        return builder

    @staticmethod
    def _do(query: AggregateBuilder) -> dict:
        res = query.do()
        if (errs := res.get("errors")) is not None:
            if "Unexpected empty IN" in errs[0]["message"]:
                raise WeaviateQueryError(
                    "The query that you sent had no body so GraphQL was unable to parse it. You must provide at least one option to the aggregation method in order to build a valid query.",
                    "GQL Aggregate",
                )
            raise WeaviateQueryError(
                f"Error in GraphQL response: {json.dumps(errs, indent=2)}, for the following query: {query.build()}",
                "GQL Aggregate",
            )
        return res

    @staticmethod
    def _parse_near_options(
        certainty: Optional[NUMBER],
        distance: Optional[NUMBER],
        object_limit: Optional[int],
    ) -> None:
        _validate_input(
            [
                _ValidateArgument([int, float, None], "certainty", certainty),
                _ValidateArgument([int, float, None], "distance", distance),
                _ValidateArgument([int, None], "object_limit", object_limit),
            ]
        )

    @staticmethod
    def _add_hybrid_to_builder(
        builder: AggregateBuilder,
        query: Optional[str],
        alpha: Optional[NUMBER],
        vector: Optional[List[float]],
        query_properties: Optional[List[str]],
        object_limit: Optional[int],
        target_vector: Optional[str],
    ) -> AggregateBuilder:
        payload: dict = {}
        if query is not None:
            payload["query"] = query
        if alpha is not None:
            payload["alpha"] = alpha
        if vector is not None:
            payload["vector"] = vector
        if query_properties is not None:
            payload["properties"] = query_properties
        if target_vector is not None:
            payload["targetVectors"] = [target_vector]
        builder = builder.with_hybrid(payload)
        if object_limit is not None:
            builder = builder.with_object_limit(object_limit)
        return builder

    @staticmethod
    def _add_near_image_to_builder(
        builder: AggregateBuilder,
        near_image: Union[str, pathlib.Path, io.BufferedReader],
        certainty: Optional[NUMBER],
        distance: Optional[NUMBER],
        object_limit: Optional[int],
        target_vector: Optional[str],
    ) -> AggregateBuilder:
        if all([certainty is None, distance is None, object_limit is None]):
            raise WeaviateInvalidInputError(
                "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
            )
        _validate_input(
            _ValidateArgument([str, pathlib.Path, io.BufferedReader], "near_image", near_image)
        )
        _Aggregate._parse_near_options(certainty, distance, object_limit)
        payload: dict = {}
        payload["image"] = _parse_media(near_image)
        if certainty is not None:
            payload["certainty"] = certainty
        if distance is not None:
            payload["distance"] = distance
        if target_vector is not None:
            payload["targetVector"] = target_vector
        builder = builder.with_near_image(payload, encode=False)
        if object_limit is not None:
            builder = builder.with_object_limit(object_limit)
        return builder

    @staticmethod
    def _add_near_object_to_builder(
        builder: AggregateBuilder,
        near_object: UUID,
        certainty: Optional[NUMBER],
        distance: Optional[NUMBER],
        object_limit: Optional[int],
        target_vector: Optional[str],
    ) -> AggregateBuilder:
        if all([certainty is None, distance is None, object_limit is None]):
            raise WeaviateInvalidInputError(
                "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
            )
        _validate_input(_ValidateArgument([UUID], "near_object", near_object))
        _Aggregate._parse_near_options(certainty, distance, object_limit)
        payload: dict = {}
        payload["id"] = str(near_object)
        if certainty is not None:
            payload["certainty"] = certainty
        if distance is not None:
            payload["distance"] = distance
        if target_vector is not None:
            payload["targetVector"] = target_vector
        builder = builder.with_near_object(payload)
        if object_limit is not None:
            builder = builder.with_object_limit(object_limit)
        return builder

    @staticmethod
    def _add_near_text_to_builder(
        builder: AggregateBuilder,
        query: Union[List[str], str],
        certainty: Optional[NUMBER],
        distance: Optional[NUMBER],
        move_to: Optional[Move],
        move_away: Optional[Move],
        object_limit: Optional[int],
        target_vector: Optional[str],
    ) -> AggregateBuilder:
        if all([certainty is None, distance is None, object_limit is None]):
            raise WeaviateInvalidInputError(
                "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
            )
        _validate_input(
            [
                _ValidateArgument([List[str], str], "query", query),
                _ValidateArgument([Move, None], "move_to", move_to),
                _ValidateArgument([Move, None], "move_away", move_away),
                _ValidateArgument([str, None], "target_vector", target_vector),
            ]
        )
        _Aggregate._parse_near_options(certainty, distance, object_limit)
        payload: dict = {}
        payload["concepts"] = query if isinstance(query, list) else [query]
        if certainty is not None:
            payload["certainty"] = certainty
        if distance is not None:
            payload["distance"] = distance
        if move_to is not None:
            payload["moveTo"] = move_to._to_gql_payload()
        if move_away is not None:
            payload["moveAwayFrom"] = move_away._to_gql_payload()
        if target_vector is not None:
            payload["targetVector"] = target_vector
        builder = builder.with_near_text(payload)
        if object_limit is not None:
            builder = builder.with_object_limit(object_limit)
        return builder

    @staticmethod
    def _add_near_vector_to_builder(
        builder: AggregateBuilder,
        near_vector: List[float],
        certainty: Optional[NUMBER],
        distance: Optional[NUMBER],
        object_limit: Optional[int],
        target_vector: Optional[str],
    ) -> AggregateBuilder:
        if all([certainty is None, distance is None, object_limit is None]):
            raise WeaviateInvalidInputError(
                "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
            )
        _validate_input(_ValidateArgument([list], "near_vector", near_vector))
        _Aggregate._parse_near_options(certainty, distance, object_limit)
        payload: dict = {}
        payload["vector"] = near_vector
        if certainty is not None:
            payload["certainty"] = certainty
        if distance is not None:
            payload["distance"] = distance
        if target_vector is not None:
            payload["targetVector"] = target_vector
        builder = builder.with_near_vector(payload)
        if object_limit is not None:
            builder = builder.with_object_limit(object_limit)
        return builder


def _parse_media(media: Union[str, pathlib.Path, io.BufferedReader]) -> str:
    if isinstance(media, str):  # if already encoded by user or string to path
        if os.path.isfile(media):
            return file_encoder_b64(media)
        else:
            return media
    else:
        return file_encoder_b64(media)



from typing import (
    Generic,
    Optional,
    cast,
)

from weaviate.collections.classes.filters import (
    Filter,
)
from weaviate.collections.classes.grpc import MetadataQuery
from weaviate.collections.classes.internal import (
    ObjectSingleReturn,
    MetadataSingleObjectReturn,
    QuerySingleReturn,
    ReturnProperties,
    ReturnReferences,
    _QueryOptions,
)
from weaviate.collections.classes.types import Properties, TProperties, References, TReferences
from weaviate.collections.queries.base import _BaseQuery
from weaviate.types import INCLUDE_VECTOR, UUID


class _FetchObjectByIDQuery(Generic[Properties, References], _BaseQuery[Properties, References]):
    def fetch_object_by_id(
        self,
        uuid: UUID,
        include_vector: INCLUDE_VECTOR = False,
        *,
        return_properties: Optional[ReturnProperties[TProperties]] = None,
        return_references: Optional[ReturnReferences[TReferences]] = None,
    ) -> Optional[QuerySingleReturn[Properties, References, TProperties, TReferences]]:
        """Retrieve an object from the server by its UUID.

        Arguments:
            `uuid`
                The UUID of the object to retrieve, REQUIRED.
            `include_vector`
                Whether to include the vector in the returned object.
            `return_properties`
                The properties to return for each object.
            `return_references`
                The references to return for each object.

        NOTE:
            - If `return_properties` is not provided then all properties are returned except for blob properties.
            - If `return_metadata` is not provided then no metadata is provided. Use MetadataQuery.full() to retrieve all metadata.
            - If `return_references` is not provided then no references are provided.

        Raises:
            `weaviate.exceptions.WeaviateGRPCQueryError`:
                If the network connection to Weaviate fails.
            `weaviate.exceptions.WeaviateInsertInvalidPropertyError`:
                If a property is invalid. I.e., has name `id` or `vector`, which are reserved.
        """
        return_metadata = MetadataQuery(
            creation_time=True, last_update_time=True, is_consistent=True
        )
        res = self._query.get(
            limit=1,
            filters=Filter.by_id().equal(uuid),
            return_metadata=self._parse_return_metadata(return_metadata, include_vector),
            return_properties=self._parse_return_properties(return_properties),
            return_references=self._parse_return_references(return_references),
        )
        objects = self._result_to_query_return(
            res,
            _QueryOptions.from_input(
                return_metadata,
                return_properties,
                include_vector,
                self._references,
                return_references,
            ),
            return_properties,
            None,
        )

        if len(objects.objects) == 0:
            return None

        obj = objects.objects[0]
        assert obj.metadata is not None
        assert obj.metadata.creation_time is not None
        assert obj.metadata.last_update_time is not None

        return cast(
            QuerySingleReturn[Properties, References, TProperties, TReferences],
            ObjectSingleReturn(
                uuid=obj.uuid,
                vector=obj.vector,
                properties=obj.properties,
                metadata=MetadataSingleObjectReturn(
                    creation_time=obj.metadata.creation_time,
                    last_update_time=obj.metadata.last_update_time,
                    is_consistent=obj.metadata.is_consistent,
                ),
                references=obj.references,
                collection=obj.collection,
            ),
        )



from typing import Optional, Sequence

from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.collections.classes.tenants import TenantActivityStatus
from weaviate.collections.grpc.shared import _BaseGRPC

from weaviate.connect import ConnectionV4

from weaviate.proto.v1 import tenants_pb2


class _TenantsGRPC(_BaseGRPC):
    def __init__(
        self,
        connection: ConnectionV4,
        name: str,
        consistency_level: Optional[ConsistencyLevel],
    ):
        super().__init__(connection, consistency_level)
        self._name: str = name

    def get(self, names: Optional[Sequence[str]]) -> tenants_pb2.TenantsGetReply:
        assert self._connection.grpc_stub is not None, "gRPC stub is not initialized"

        request = tenants_pb2.TenantsGetRequest(
            collection=self._name,
            names=tenants_pb2.TenantNames(values=names) if names is not None else None,
        )
        res: tenants_pb2.TenantsGetReply  # According to PEP-0526
        res, _ = self._connection.grpc_stub.TenantsGet.with_call(
            request,
            metadata=self._connection.grpc_headers(),
            timeout=self._connection.timeout_config.query,
        )
        return res

    def map_activity_status(self, status: tenants_pb2.TenantActivityStatus) -> TenantActivityStatus:
        if status == tenants_pb2.TENANT_ACTIVITY_STATUS_COLD:
            return TenantActivityStatus.COLD
        if status == tenants_pb2.TENANT_ACTIVITY_STATUS_HOT:
            return TenantActivityStatus.HOT
        if status == tenants_pb2.TENANT_ACTIVITY_STATUS_FROZEN:
            return TenantActivityStatus.FROZEN
        if status == tenants_pb2.TENANT_ACTIVITY_STATUS_WARM:
            return TenantActivityStatus.WARM
        raise ValueError(f"Unknown TenantActivityStatus: {status}")



import pytest

from weaviate.util import _ServerVersion


@pytest.mark.parametrize(
    "version_string,expected",
    [
        ("1.2.3.a", _ServerVersion(1, 2, 3)),
        ("1.2.3.4", _ServerVersion(1, 2, 3)),
        ("v1.2.3.4", _ServerVersion(1, 2, 3)),
        ("1.2.3", _ServerVersion(1, 2, 3)),
        ("v1.2.3", _ServerVersion(1, 2, 3)),
        ("1.2", _ServerVersion(1, 2, 0)),
        ("v1.2", _ServerVersion(1, 2, 0)),
        ("1", _ServerVersion(1, 0, 0)),
        ("v1", _ServerVersion(1, 0, 0)),
        ("", _ServerVersion(0, 0, 0)),
    ],
)
def test_server_version_successful_parsing(version_string: str, expected: _ServerVersion) -> None:
    assert _ServerVersion.from_string(version_string) == expected


@pytest.mark.parametrize(
    "version_string",
    [
        "v",
        "1.v2.3",
        "1.2.b3",
    ],
)
def test_server_version_unsuccessful_parsing(version_string: str) -> None:
    with pytest.raises(ValueError):
        _ServerVersion.from_string(version_string)


@pytest.mark.parametrize(
    "is_valid",
    [
        _ServerVersion(1, 2, 3).is_at_least(1, 2, 3),
        _ServerVersion(1, 2, 3).is_at_least(1, 2, 2),
        _ServerVersion(1, 2, 3).is_at_least(1, 1, 3),
        _ServerVersion(1, 2, 3).is_at_least(1, 1, 2),
        _ServerVersion(1, 2, 3).is_at_least(0, 2, 3),
        _ServerVersion(1, 2, 3).is_at_least(0, 2, 2),
        _ServerVersion(1, 2, 3).is_at_least(0, 1, 3),
        _ServerVersion(1, 2, 3).is_at_least(0, 1, 2),
        not _ServerVersion(1, 2, 3).is_at_least(1, 2, 4),
        not _ServerVersion(1, 2, 3).is_at_least(1, 3, 3),
        not _ServerVersion(1, 2, 3).is_at_least(1, 3, 4),
        not _ServerVersion(1, 2, 3).is_at_least(2, 2, 3),
        not _ServerVersion(1, 2, 3).is_at_least(2, 2, 4),
        not _ServerVersion(1, 2, 3).is_at_least(2, 3, 3),
        not _ServerVersion(1, 2, 3).is_at_least(2, 3, 4),
        not _ServerVersion(0, 0, 0).is_at_least(0, 0, 1),
    ],
)
def test_server_version_is_at_least(is_valid: bool) -> None:
    assert is_valid


def test_server_version_magic_methods() -> None:
    # Test __eq__
    assert _ServerVersion(1, 2, 3) == _ServerVersion(1, 2, 3)
    # Test __neq__
    assert _ServerVersion(1, 2, 3) != _ServerVersion(1, 2, 2)

    # Test __lt__
    assert _ServerVersion(1, 2, 2) < _ServerVersion(1, 2, 3)
    assert _ServerVersion(1, 1, 3) < _ServerVersion(1, 2, 3)
    assert _ServerVersion(0, 2, 3) < _ServerVersion(1, 2, 3)

    # Test __le__
    assert _ServerVersion(1, 2, 3) <= _ServerVersion(1, 2, 3)
    assert _ServerVersion(1, 2, 2) <= _ServerVersion(1, 2, 3)
    assert _ServerVersion(1, 1, 3) <= _ServerVersion(1, 2, 3)
    assert _ServerVersion(0, 2, 3) <= _ServerVersion(1, 2, 3)

    # Test __gt__
    assert _ServerVersion(1, 2, 4) > _ServerVersion(1, 2, 3)
    assert _ServerVersion(1, 3, 3) > _ServerVersion(1, 2, 3)
    assert _ServerVersion(2, 2, 3) > _ServerVersion(1, 2, 3)

    # Test __ge__
    assert _ServerVersion(1, 2, 3) >= _ServerVersion(1, 2, 3)
    assert _ServerVersion(1, 2, 4) >= _ServerVersion(1, 2, 3)
    assert _ServerVersion(1, 3, 3) >= _ServerVersion(1, 2, 3)
    assert _ServerVersion(2, 2, 3) >= _ServerVersion(1, 2, 3)



"""
Helper functions!
"""

import base64
import datetime
import io
import json
import os
import re
from enum import Enum, EnumMeta
from pathlib import Path
from typing import Union, Sequence, Any, Optional, List, Dict, Generator, Tuple, cast

import requests
import httpx
import uuid as uuid_lib
import validators
from requests.exceptions import JSONDecodeError

from weaviate.exceptions import (
    SchemaValidationError,
    UnexpectedStatusCodeError,
    ResponseCannotBeDecodedError,
    WeaviateInvalidInputError,
    WeaviateUnsupportedFeatureError,
)
from weaviate.warnings import _Warnings
from weaviate.types import NUMBER, UUIDS, TIME

PYPI_PACKAGE_URL = "https://pypi.org/pypi/weaviate-client/json"
MAXIMUM_MINOR_VERSION_DELTA = 3  # The maximum delta between minor versions of Weaviate Client that will not trigger an upgrade warning.
MINIMUM_NO_WARNING_VERSION = (
    "v1.16.0"  # The minimum version of Weaviate that will not trigger an upgrade warning.
)
BYTES_PER_CHUNK = 65535  # The number of bytes to read per chunk when encoding files ~ 64kb


# MetaEnum and BaseEnum are required to support `in` statements:
#    'ALL' in ConsistencyLevel == True
#    12345 in ConsistencyLevel == False
class MetaEnum(EnumMeta):
    def __contains__(cls, item: Any) -> bool:
        try:
            # when item is type ConsistencyLevel
            return item.name in cls.__members__.keys()
        except AttributeError:
            # when item is type str
            return item in cls.__members__.keys()


class BaseEnum(Enum, metaclass=MetaEnum):
    pass


def image_encoder_b64(image_or_image_path: Union[str, io.BufferedReader]) -> str:
    """
    Encode a image in a Weaviate understandable format from a binary read file or by providing
    the image path.

    Parameters
    ----------
    image_or_image_path : str, io.BufferedReader
        The binary read file or the path to the file.

    Returns
    -------
    str
        Encoded image.

    Raises
    ------
    ValueError
        If the argument is str and does not point to an existing file.
    TypeError
        If the argument is of a wrong data type.
    """

    if isinstance(image_or_image_path, str):
        if not os.path.isfile(image_or_image_path):
            raise ValueError("No file found at location " + image_or_image_path)
        with open(image_or_image_path, "br") as file:
            content = file.read()

    elif isinstance(image_or_image_path, io.BufferedReader):
        content = image_or_image_path.read()
    else:
        raise TypeError(
            '"image_or_image_path" should be a image path or a binary read file'
            " (io.BufferedReader)"
        )
    return base64.b64encode(content).decode("utf-8")


def file_encoder_b64(file_or_file_path: Union[str, Path, io.BufferedReader]) -> str:
    """
    Encode a file in a Weaviate understandable format from an io.BufferedReader binary read file or by providing
    the file path as either a string of a pathlib.Path object

    If you pass an io.BufferedReader object, it is your responsibility to close it after encoding.

    Parameters
    ----------
    file_or_file_path : str, pathlib.Path io.BufferedReader
        The binary read file or the path to the file.

    Returns
    -------
    str
        Encoded file.

    Raises
    ------
    ValueError
        If the argument is str and does not point to an existing file.
    TypeError
        If the argument is of a wrong data type.
    """

    def _chunks(buffer: io.BufferedReader, chunk_size: int) -> Generator[bytes, Any, Any]:
        while True:
            data = buffer.read(chunk_size)
            if not data:
                break
            yield data

    should_close_file = False
    use_buffering = True
    file = None

    try:
        if isinstance(file_or_file_path, str):
            if not os.path.isfile(file_or_file_path):
                raise ValueError("No file found at location " + file_or_file_path)
            file = open(file_or_file_path, "br")
            should_close_file = True
            use_buffering = os.path.getsize(file_or_file_path) > BYTES_PER_CHUNK
        elif isinstance(file_or_file_path, Path):
            if not file_or_file_path.is_file():
                raise ValueError("No file found at location " + str(file_or_file_path))
            file = file_or_file_path.open("br")
            should_close_file = True
            use_buffering = file_or_file_path.stat().st_size > BYTES_PER_CHUNK
        elif isinstance(file_or_file_path, io.BufferedReader):
            file = file_or_file_path
        else:
            raise TypeError(
                '"file_or_file_path" should be a file path or a binary read file'
                " (io.BufferedReader)"
            )

        if use_buffering:
            encoded: str = ""
            for chunk in _chunks(file, BYTES_PER_CHUNK):
                encoded += base64.b64encode(chunk).decode("utf-8")
        else:
            encoded = base64.b64encode(file.read()).decode("utf-8")

    finally:
        if should_close_file and file is not None:
            file.close()

    return encoded


def image_decoder_b64(encoded_image: str) -> bytes:
    """
    Decode image from a Weaviate format image.

    Parameters
    ----------
    encoded_image : str
        The encoded image.

    Returns
    -------
    bytes
        Decoded image as a binary string.
    """

    return base64.b64decode(encoded_image.encode("utf-8"))


def file_decoder_b64(encoded_file: str) -> bytes:
    """
    Decode file from a Weaviate format image.

    Parameters
    ----------
    encoded_file : str
        The encoded file.

    Returns
    -------
    bytes
        Decoded file as a binary string. Use this in your file
        handling code to convert it into a specific file type of choice.
        E.g., PIL for images.
    """

    return base64.b64decode(encoded_file.encode("utf-8"))


def generate_local_beacon(
    to_uuid: Union[str, uuid_lib.UUID],
    class_name: Optional[str] = None,
) -> dict:
    """
    Generates a beacon with the given uuid and class name (only for Weaviate >= 1.14.0).

    Parameters
    ----------
    to_uuid : str or uuid.UUID
        The UUID for which to create a local beacon.
    class_name : Optional[str], optional
        The class name of the `to_uuid` object. Used with Weaviate >= 1.14.0.
        For Weaviate < 1.14.0 use None value.

    Returns
    -------
    dict
        The local beacon.

    Raises
    ------
    TypeError
        If 'to_uuid' is not of type str.
    ValueError
        If the 'to_uuid' is not valid.
    """

    if isinstance(to_uuid, str):
        try:
            uuid = str(uuid_lib.UUID(to_uuid))
        except ValueError:
            raise ValueError("Uuid does not have the proper form") from None
    elif isinstance(to_uuid, uuid_lib.UUID):
        uuid = str(to_uuid)
    else:
        raise TypeError("Expected to_object_uuid of type str or uuid.UUID")

    if class_name is None:
        return {"beacon": f"weaviate://localhost/{uuid}"}
    return {"beacon": f"weaviate://localhost/{_capitalize_first_letter(class_name)}/{uuid}"}


def _get_dict_from_object(object_: Union[str, dict]) -> dict:
    """
    Takes an object that should describe a dict
    e.g. a schema or an object and tries to retrieve the dict.

    Parameters
    ----------
    object_ : str or dict
        The object from which to retrieve the dict.
        Can be a python dict, or the path to a json file or a url of a json file.

    Returns
    -------
    dict
        The object as a dict.

    Raises
    ------
    TypeError
        If 'object_' is neither a string nor a dict.
    ValueError
        If no dict can be retrieved from object.
    """

    # check if objects files is url
    if object_ is None:
        raise TypeError("argument is None")

    if isinstance(object_, dict):
        # Object is already a dict
        return object_
    if isinstance(object_, str):
        if validators.url(object_):
            # Object is URL
            response = requests.get(object_)
            if response.status_code == 200:
                return cast(dict, response.json())
            raise ValueError("Could not download file " + object_)

        if not os.path.isfile(object_):
            # Object is neither file nor URL
            raise ValueError("No file found at location " + object_)
        # Object is file
        with open(object_, "r") as file:
            return cast(dict, json.load(file))
    raise TypeError(
        "Argument is not of the supported types. Supported types are "
        "url or file path as string or schema as dict."
    )


def is_weaviate_object_url(url: str) -> bool:
    """
    Checks if the input follows a normal Weaviate 'beacon' like this:
    'weaviate://localhost/ClassName/28f3f61b-b524-45e0-9bbe-2c1550bf73d2'

    Parameters
    ----------
    url : str
        The URL to be validated.

    Returns
    -------
    bool
        True if the 'url' is a Weaviate object URL.
        False otherwise.
    """

    if not isinstance(url, str):
        return False
    if not url.startswith("weaviate://"):
        return False
    url = url[11:]
    split = url.split("/")
    if len(split) not in (2, 3):
        return False
    if split[0] != "localhost":
        if not validators.domain(split[0]):
            return False
    try:
        uuid_lib.UUID(split[-1])
    except ValueError:
        return False
    return True


def is_object_url(url: str) -> bool:
    """
    Validates an url like 'http://localhost:8080/v1/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446'
    or '/v1/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446' references a object. It only validates
    the path format and UUID, not the host or the protocol.

    Parameters
    ----------
    url : str
        The URL to be validated.

    Returns
    -------
    bool
        True if the 'url' is a valid path to an object.
        False otherwise.
    """

    v1_split = url.split("/v1/")

    if len(v1_split) != 2:
        return False

    split = v1_split[1].split("/")

    if len(split) not in (2, 3):
        return False

    try:
        uuid_lib.UUID(split[-1])
    except ValueError:
        return False
    if not split[0] == "objects":
        return False
    return True


def get_valid_uuid(uuid: Union[str, uuid_lib.UUID]) -> str:
    """
    Validate and extract the UUID.

    Parameters
    ----------
    uuid : str or uuid.UUID
        The UUID to be validated and extracted.
        Should be in the form of an UUID or in form of an URL (weaviate 'beacon' or 'href').
        E.g.
        'http://localhost:8080/v1/objects/fc7eb129-f138-457f-b727-1b29db191a67'
        or
        'weaviate://localhost/28f3f61b-b524-45e0-9bbe-2c1550bf73d2'
        or
        'fc7eb129-f138-457f-b727-1b29db191a67'

    Returns
    -------
    str
        The extracted UUID.

    Raises
    ------
    TypeError
        If 'uuid' is not of type str.
    ValueError
        If 'uuid' is not valid or cannot be extracted.
    """

    if isinstance(uuid, uuid_lib.UUID):
        return str(uuid)

    if not isinstance(uuid, str):
        raise TypeError("'uuid' must be of type str or uuid.UUID, but was: " + str(type(uuid)))

    _is_weaviate_url = is_weaviate_object_url(uuid)
    _is_object_url = is_object_url(uuid)
    _uuid = uuid
    if _is_weaviate_url or _is_object_url:
        _uuid = uuid.split("/")[-1]
    try:
        _uuid = str(uuid_lib.UUID(_uuid))
    except ValueError:
        raise ValueError("Not valid 'uuid' or 'uuid' can not be extracted from value") from None
    return _uuid


def get_vector(vector: Sequence) -> List[float]:
    """
    Get weaviate compatible format of the embedding vector.

    Parameters
    ----------
    vector: Sequence
        The embedding of an object. Used only for class objects that do not have a vectorization
        module. Supported types are `list`, `numpy.ndarray`, `torch.Tensor`, `tf.Tensor`, `pd.Series` and `pl.Series`.

    Returns
    -------
    list
        The embedding as a list.

    Raises
    ------
    TypeError
        If 'vector' is not of a supported type.
    """

    if isinstance(vector, list):
        # if vector is already a list
        return vector
    try:
        # if vector is numpy.ndarray or torch.Tensor
        return vector.squeeze().tolist()  # type: ignore
    except AttributeError:
        pass
    try:
        # if vector is tf.Tensor or torch.Tensor
        return vector.numpy().squeeze().tolist()  # type: ignore
    except AttributeError:
        pass
    try:
        # if vector is pd.Series or pl.Series
        return vector.to_list()  # type: ignore
    except AttributeError:
        pass
    raise TypeError(
        "The type of the 'vector' argument is not supported!\n"
        "Supported types are `list`, 'numpy.ndarray`, `torch.Tensor`, `tf.Tensor`, `pd.Series`, and `pl.Series`"
    ) from None


def _get_vector_v4(vector: Sequence) -> List[float]:
    try:
        return get_vector(vector)
    except TypeError as e:
        raise WeaviateInvalidInputError(
            f"The vector you supplied was malformatted! Vector:  {vector}"
        ) from e


def get_domain_from_weaviate_url(url: str) -> str:
    """
    Get the domain from a weaviate URL.

    Parameters
    ----------
    url : str
        The weaviate URL.
        Of this form: 'weaviate://localhost/objects/28f3f61b-b524-45e0-9bbe-2c1550bf73d2'

    Returns
    -------
    str
        The domain.
    """

    return url[11:].split("/")[0]


def _is_sub_schema(sub_schema: dict, schema: dict) -> bool:
    """
    Check for a subset in a schema.

    Parameters
    ----------
    sub_schema : dict
        The smaller schema that should be contained in the 'schema'.
    schema : dict
        The schema for which to check if 'sub_schema' is a part of. Must have the 'classes' key.

    Returns
    -------
    bool
        True is 'sub_schema' is a subset of the 'schema'.
        False otherwise.
    """

    schema_classes = schema.get("classes", [])
    if "classes" in sub_schema:
        sub_schema_classes = sub_schema["classes"]
    else:
        sub_schema_classes = [sub_schema]
    return _compare_class_sets(sub_schema_classes, schema_classes)


def _compare_class_sets(sub_set: list, set_: list) -> bool:
    """
    Check for a subset in a set of classes.

    Parameters
    ----------
    sub_set : list
        The smaller set that should be contained in the 'set'.
    set_ : list
        The set for which to check if 'sub_set' is a part of.

    Returns
    -------
    bool
        True is 'sub_set' is a subset of the 'set'.
        False otherwise.
    """

    for sub_set_class in sub_set:
        found = False
        for set_class in set_:
            if "class" not in sub_set_class:
                raise SchemaValidationError(
                    "The sub schema class/es MUST have a 'class' keyword each!"
                )
            if _capitalize_first_letter(sub_set_class["class"]) == _capitalize_first_letter(
                set_class["class"]
            ):
                if _compare_properties(sub_set_class["properties"], set_class["properties"]):
                    found = True
                    break
        if not found:
            return False
    return True


def _compare_properties(sub_set: list, set_: list) -> bool:
    """
    Check for a subset in a set of properties.

    Parameters
    ----------
    sub_set : list
        The smaller set that should be contained in the 'set'.
    set_ : list
        The set for which to check if 'sub_set' is a part of.

    Returns
    -------
    bool
        True is 'sub_set' is a subset of the 'set'.
        False otherwise.
    """

    for sub_set_property in sub_set:
        found = False
        for set_property in set_:
            if sub_set_property["name"] == set_property["name"]:
                found = True
                break
        if not found:
            return False
    return True


def generate_uuid5(identifier: Any, namespace: Any = "") -> str:
    """
    Generate an UUIDv5, may be used to consistently generate the same UUID for a specific
    identifier and namespace.

    Parameters
    ----------
    identifier : Any
        The identifier/object that should be used as basis for the UUID.
    namespace : Any, optional
        Allows to namespace the identifier, by default ""

    Returns
    -------
    str
        The UUID as a string.
    """

    return str(uuid_lib.uuid5(uuid_lib.NAMESPACE_DNS, str(namespace) + str(identifier)))


def _capitalize_first_letter(string: str) -> str:
    """
    Capitalize only the first letter of the `string`.

    Parameters
    ----------
    string : str
        The string to be capitalized.

    Returns
    -------
    str
        The capitalized string.
    """

    if len(string) == 1:
        return string.capitalize()
    return string[0].capitalize() + string[1:]


def check_batch_result(
    results: Optional[List[Dict[str, Any]]],
) -> None:
    """
    Check batch results for errors.

    Parameters
    ----------
    results : dict
        The Weaviate batch creation return value.
    """

    if results is None:
        return
    for result in results:
        if "result" in result and "errors" in result["result"]:
            if "error" in result["result"]["errors"]:
                print(result["result"]["errors"])


def _check_positive_num(
    value: Any, arg_name: str, data_type: type, include_zero: bool = False
) -> None:
    """
    Check if the `value` of the `arg_name` is a positive number.

    Parameters
    ----------
    value : Union[int, float]
        The value to check.
    arg_name : str
        The name of the variable from the original function call. Used for error message.
    data_type : type
        The data type to check for.
    include_zero : bool
        Wether zero counts as positive or not. By default False.

    Raises
    ------
    TypeError
        If the `value` is not of type `data_type`.
    ValueError
        If the `value` has a non positive value.
    """

    if not isinstance(value, data_type) or isinstance(value, bool):
        raise TypeError(f"'{arg_name}' must be of type {data_type}.")
    if include_zero:
        if value < 0:  # type: ignore
            raise ValueError(f"'{arg_name}' must be positive, i.e. greater or equal to zero (>=0).")
    else:
        if value <= 0:  # type: ignore
            raise ValueError(f"'{arg_name}' must be positive, i.e. greater that zero (>0).")


def is_weaviate_domain(url: str) -> bool:
    return (
        "weaviate.io" in url.lower()
        or "semi.technology" in url.lower()
        or "weaviate.cloud" in url.lower()
    )


def strip_newlines(s: str) -> str:
    return s.replace("\n", " ")


def _sanitize_str(value: str) -> str:
    """
    Ensures string is sanitized for GraphQL.

    Parameters
    ----------
    value : str
        The value to be converted.

    Returns
    -------
    str
        The sanitized string.
    """
    value = strip_newlines(value)
    value = re.sub(r'(?<!\\)"', '\\"', value)  # only replaces unescaped double quotes
    return f'"{value}"'


def parse_version_string(ver_str: str) -> tuple:
    """
    Parse a version string into a float.

    Parameters
    ----------
    ver_str : str
        The version string to parse. (e.g. "v1.18.2" or "1.18.0")

    Returns
    -------
    tuple :
        The parsed version as a tuple with len(2). (e.g. (1, 18)) Note: Ignores the patch version.
    """
    if ver_str.count(".") == 0:
        ver_str = ver_str + ".0"

    pattern = r"v?(\d+)\.(\d+)"
    match = re.match(pattern, ver_str)

    if match:
        ver_tup = tuple(map(int, match.groups()))
        return ver_tup
    else:
        raise ValueError(
            f"Unable to parse a version from the input string: {ver_str}. Is it in the format '(v)x.y.z' (e.g. 'v1.18.2' or '1.18.0')?"
        )


class _ServerVersion:
    def __init__(self, major: int, minor: int, patch: int) -> None:
        self.major = major
        self.minor = minor
        self.patch = patch

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _ServerVersion):
            return NotImplemented
        return self.major == other.major and self.minor == other.minor and self.patch == other.patch

    def __neq__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __gt__(self, other: "_ServerVersion") -> bool:
        if self.major > other.major:
            return True
        elif self.major == other.major:
            if self.minor > other.minor:
                return True
            elif self.minor == other.minor:
                if self.patch > other.patch:
                    return True
        return False

    def __lt__(self, other: "_ServerVersion") -> bool:
        return not self.__gt__(other) and not self.__eq__(other)

    def __ge__(self, other: "_ServerVersion") -> bool:
        return self.__gt__(other) or self.__eq__(other)

    def __le__(self, other: "_ServerVersion") -> bool:
        return self.__lt__(other) or self.__eq__(other)

    def __repr__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"

    def is_at_least(self, major: int, minor: int, patch: int) -> bool:
        return self >= _ServerVersion(major, minor, patch)

    def is_lower_than(self, major: int, minor: int, patch: int) -> bool:
        return self < _ServerVersion(major, minor, patch)

    @classmethod
    def from_string(cls, version: str) -> "_ServerVersion":
        initial = version
        if version == "":
            version = "0"
        if version.count(".") == 0:
            version = version + ".0"
        if version.count(".") == 1:
            version = version + ".0"

        pattern = r"v?(\d+)\.(\d+)\.(\d+)"
        match = re.match(pattern, version)

        if match:
            ver_tup = tuple(map(int, match.groups()))
            return cls(major=ver_tup[0], minor=ver_tup[1], patch=ver_tup[2])
        else:
            raise ValueError(
                f"Unable to parse a version from the input string: {initial}. Is it in the format '(v)x.y.z' (e.g. 'v1.18.2' or '1.18.0')?"
            )

    def check_is_at_least_1_25_0(self, feature: str) -> None:
        if not self >= _ServerVersion(1, 25, 0):
            raise WeaviateUnsupportedFeatureError(feature, str(self), "1.25.0")

    @property
    def supports_tenants_get_grpc(self) -> bool:
        return self >= _ServerVersion(1, 25, 0)


def is_weaviate_too_old(current_version_str: str) -> bool:
    """
    Check if the user should be gently nudged to upgrade their Weaviate server version.

    Parameters
    ----------
    current_version_str : str
        The version of the Weaviate server that the client is connected to. (e.g. "v1.18.2" or "1.18.0")

    Returns
    -------
    bool :
    True if the user should be nudged to upgrade.

    """

    current_version = parse_version_string(current_version_str)
    minimum_version = parse_version_string(MINIMUM_NO_WARNING_VERSION)
    return minimum_version > current_version


def is_weaviate_client_too_old(current_version_str: str, latest_version_str: str) -> bool:
    """
    Check if the user should be gently nudged to upgrade their Weaviate client version.

    Parameters
    ----------
    current_version_str : str
        The version of the Weaviate client that is being used (e.g. "v1.18.2" or "1.18.0")
    latest_version_str : str
        The latest version of the Weaviate client to compare against (e.g. "v1.18.2" or "1.18.0")

    Returns
    -------
    bool :
    True if the user should be nudged to upgrade.
    False if the user is using a valid version or if the version could not be parsed.

    """

    try:
        current_version = parse_version_string(current_version_str)
        latest_major, latest_minor = parse_version_string(latest_version_str)
        minimum_minor = max(latest_minor - MAXIMUM_MINOR_VERSION_DELTA, 0)
        minimum_version = (latest_major, minimum_minor)
        return minimum_version > current_version
    except ValueError:
        return False


def _get_valid_timeout_config(
    timeout_config: Union[Tuple[NUMBER, NUMBER], NUMBER, None]
) -> Tuple[NUMBER, NUMBER]:
    """
    Validate and return TimeOut configuration.

    Parameters
    ----------
    timeout_config : tuple(NUMBERS, NUMBERS) or NUMBERS or None, optional
            Set the timeout configuration for all requests to the Weaviate server. It can be a
            number or, a tuple of two numbers: (connect timeout, read timeout).
            If only one number is passed then both connect and read timeout will be set to
            that value.

    Raises
    ------
    TypeError
        If arguments are of a wrong data type.
    ValueError
        If 'timeout_config' is not a tuple of 2.
    ValueError
        If 'timeout_config' is/contains negative number/s.
    """

    def check_number(num: Union[NUMBER, Tuple[NUMBER, NUMBER], None]) -> bool:
        return isinstance(num, float) or isinstance(num, int)

    if (isinstance(timeout_config, float) or isinstance(timeout_config, int)) and not isinstance(
        timeout_config, bool
    ):
        assert timeout_config is not None
        if timeout_config <= 0.0:
            raise ValueError("'timeout_config' cannot be non-positive number/s!")
        return timeout_config, timeout_config

    if not isinstance(timeout_config, tuple):
        raise TypeError("'timeout_config' should be a (or tuple of) positive number/s!")
    if len(timeout_config) != 2:
        raise ValueError("'timeout_config' must be of length 2!")
    if not (check_number(timeout_config[0]) and check_number(timeout_config[1])) or (
        isinstance(timeout_config[0], bool) and isinstance(timeout_config[1], bool)
    ):
        raise TypeError("'timeout_config' must be tuple of numbers")
    if timeout_config[0] <= 0.0 or timeout_config[1] <= 0.0:
        raise ValueError("'timeout_config' cannot be non-positive number/s!")
    return timeout_config


def _type_request_response(json_response: Any) -> Optional[Dict[str, Any]]:
    if json_response is None:
        return None
    assert isinstance(json_response, dict)
    return json_response


def _to_beacons(uuids: UUIDS, to_class: str = "") -> List[Dict[str, str]]:
    if isinstance(uuids, uuid_lib.UUID) or isinstance(
        uuids, str
    ):  # replace with isinstance(uuids, UUID) in 3.10
        uuids = [uuids]

    if len(to_class) > 0:
        to_class = to_class + "/"

    return [{"beacon": f"weaviate://localhost/{to_class}{uuid_to}"} for uuid_to in uuids]


def _decode_json_response_dict(
    response: Union[httpx.Response, requests.Response], location: str
) -> Optional[Dict[str, Any]]:
    if response is None:
        return None

    if 200 <= response.status_code < 300:
        try:
            json_response = cast(Dict[str, Any], response.json())
            return json_response
        except JSONDecodeError:
            raise ResponseCannotBeDecodedError(location, response)

    raise UnexpectedStatusCodeError(location, response)


def _decode_json_response_list(
    response: Union[httpx.Response, requests.Response], location: str
) -> Optional[List[Dict[str, Any]]]:
    if response is None:
        return None

    if 200 <= response.status_code < 300:
        try:
            json_response = response.json()
            return cast(list, json_response)
        except JSONDecodeError:
            raise ResponseCannotBeDecodedError(location, response)
    raise UnexpectedStatusCodeError(location, response)


def _datetime_to_string(value: TIME) -> str:
    if value.tzinfo is None:
        _Warnings.datetime_insertion_with_no_specified_timezone(value)
        value = value.replace(tzinfo=datetime.timezone.utc)
    return value.isoformat(sep="T", timespec="microseconds")


def _datetime_from_weaviate_str(string: str) -> datetime.datetime:
    try:
        return datetime.datetime.strptime(
            "".join(string.rsplit(":", 1) if string[-1] != "Z" else string),
            "%Y-%m-%dT%H:%M:%S.%f%z",
        )
    except ValueError:  # if the string does not have microseconds
        return datetime.datetime.strptime(
            "".join(string.rsplit(":", 1) if string[-1] != "Z" else string),
            "%Y-%m-%dT%H:%M:%S%z",
        )



from typing import Any, Dict, List, Optional, Sequence, Union

from weaviate.collections.classes.tenants import Tenant, TenantActivityStatus
from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.collections.grpc.tenants import _TenantsGRPC
from weaviate.connect import ConnectionV4
from weaviate.validator import _validate_input, _ValidateArgument

from weaviate.connect.v4 import _ExpectedStatusCodes


class _Tenants:
    """Represents all the CRUD methods available on a collection's multi-tenancy specification within Weaviate.

    The collection must have been created with multi-tenancy enabled in order to use any of these methods. This class
    should not be instantiated directly, but is available as a property of the `Collection` class under
    the `collection.tenants` class attribute.
    """

    def __init__(
        self,
        connection: ConnectionV4,
        name: str,
        consistency_level: Optional[ConsistencyLevel] = None,
        validate_arguments: bool = True,
    ) -> None:
        self.__connection = connection
        self.__name = name
        self.__grpc = _TenantsGRPC(
            connection=connection,
            name=name,
            consistency_level=consistency_level,
        )
        self.__validate_arguments = validate_arguments

    def create(self, tenants: Union[str, Tenant, Sequence[Union[str, Tenant]]]) -> None:
        """Create the specified tenants for a collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Arguments:
            `tenants`
                A tenant name, `wvc.config.tenants.Tenant` object, or a list of tenants names
                and/or `wvc.config.tenants.Tenant` objects to add to the given collection.
                If a string is provided, the tenant will be added with the default activity status of `HOT`.

        Raises:
            `weaviate.WeaviateConnectionError`
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeError`
                If Weaviate reports a non-OK status.
            `weaviate.WeaviateInvalidInputError`
                If `tenants` is not a list of `wvc.Tenant` objects.
        """
        if self.__validate_arguments:
            _validate_input(
                [
                    _ValidateArgument(
                        expected=[str, Tenant, Sequence[Union[str, Tenant]]],
                        name="tenants",
                        value=tenants,
                    )
                ]
            )

        loaded_tenants = (
            [
                tenant.model_dump()
                if isinstance(tenant, Tenant)
                else {"name": tenant, "activityStatus": TenantActivityStatus.HOT}
                for tenant in tenants
            ]
            if isinstance(tenants, Sequence)
            else [
                {"name": tenants, "activityStatus": TenantActivityStatus.HOT}
                if isinstance(tenants, str)
                else tenants.model_dump()
            ]
        )

        path = "/schema/" + self.__name + "/tenants"
        self.__connection.post(
            path=path,
            weaviate_object=loaded_tenants,
            error_msg=f"Collection tenants may not have been added properly for {self.__name}",
            status_codes=_ExpectedStatusCodes(
                ok_in=200, error=f"Add collection tenants for {self.__name}"
            ),
        )

    def remove(self, tenants: Union[str, Tenant, Sequence[Union[str, Tenant]]]) -> None:
        """Remove the specified tenants from a collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Arguments:
            `tenants`
                A tenant name, `wvc.config.tenants.Tenant` object, or a list of tenants names
                and/or `wvc.config.tenants.Tenant` objects to remove from the given class.

        Raises:
            `weaviate.WeaviateConnectionError`
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeError`
                If Weaviate reports a non-OK status.
            `weaviate.WeaviateInvalidInputError`
                If `tenants` is not a list of strings.
        """
        if self.__validate_arguments:
            _validate_input(
                [
                    _ValidateArgument(
                        expected=[str, Tenant, Sequence[Union[str, Tenant]]],
                        name="tenants",
                        value=tenants,
                    )
                ]
            )

        loaded_tenants = (
            [tenant.name if isinstance(tenant, Tenant) else tenant for tenant in tenants]
            if isinstance(tenants, Sequence)
            else [tenants if isinstance(tenants, str) else tenants.name]
        )

        path = "/schema/" + self.__name + "/tenants"
        self.__connection.delete(
            path=path,
            weaviate_object=loaded_tenants,
            error_msg=f"Collection tenants may not have been deleted for {self.__name}",
            status_codes=_ExpectedStatusCodes(
                ok_in=200, error=f"Delete collection tenants for {self.__name}"
            ),
        )

    def __get_with_rest(self) -> Dict[str, Tenant]:
        path = "/schema/" + self.__name + "/tenants"
        response = self.__connection.get(
            path=path,
            error_msg=f"Could not get collection tenants for {self.__name}",
            status_codes=_ExpectedStatusCodes(
                ok_in=200, error=f"Get collection tenants for {self.__name}"
            ),
        )

        tenant_resp: List[Dict[str, Any]] = response.json()
        return {tenant["name"]: Tenant(**tenant) for tenant in tenant_resp}

    def __get_with_grpc(
        self, tenants: Optional[Sequence[Union[str, Tenant]]] = None
    ) -> Dict[str, Tenant]:
        response = self.__grpc.get(
            names=[tenant.name if isinstance(tenant, Tenant) else tenant for tenant in tenants]
            if tenants is not None
            else tenants
        )

        return {
            tenant.name: Tenant(
                name=tenant.name,
                activity_status=self.__grpc.map_activity_status(tenant.activity_status),
            )
            for tenant in response.tenants
        }

    def get(self) -> Dict[str, Tenant]:
        """Return all tenants currently associated with a collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Raises:
            `weaviate.WeaviateConnectionError`
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeError`
                If Weaviate reports a non-OK status.
        """
        if self.__connection._weaviate_version.supports_tenants_get_grpc:
            return self.__get_with_grpc()
        else:
            return self.__get_with_rest()

    def get_by_names(self, tenants: Sequence[Union[str, Tenant]]) -> Dict[str, Tenant]:
        """Return named tenants currently associated with a collection in Weaviate.

        If the tenant does not exist, it will not be included in the response.
        If no names are provided, all tenants will be returned.
        The collection must have been created with multi-tenancy enabled.

        Arguments:
            `tenant`
                Sequence of tenant names of wvc.tenants.Tenant objects to retrieve. To retrieve all tenants, use the `get` method.

        Raises:
            `weaviate.WeaviateConnectionError`
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeError`
                If Weaviate reports a non-OK status.
        """
        self.__connection._weaviate_version.check_is_at_least_1_25_0("The 'get_by_names' method")
        if self.__validate_arguments:
            _validate_input(
                _ValidateArgument(
                    expected=[Sequence[Union[str, Tenant]]], name="names", value=tenants
                )
            )
        return self.__get_with_grpc(tenants=tenants)

    def get_by_name(self, tenant: Union[str, Tenant]) -> Optional[Tenant]:
        """Return a specific tenant associated with a collection in Weaviate.

        If the tenant does not exist, `None` will be returned.

        The collection must have been created with multi-tenancy enabled.

        Arguments:
            `name`
                The name of the tenant to retrieve.

        Raises:
            `weaviate.WeaviateConnectionError`
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeError`
                If Weaviate reports a non-OK status.
        """
        self.__connection._weaviate_version.check_is_at_least_1_25_0("The 'get_by_name' method")
        response = self.__grpc.get(names=[tenant.name if isinstance(tenant, Tenant) else tenant])
        if len(response.tenants) == 0:
            return None
        return Tenant(
            name=response.tenants[0].name,
            activity_status=self.__grpc.map_activity_status(response.tenants[0].activity_status),
        )

    def update(self, tenants: Union[Tenant, List[Tenant]]) -> None:
        """Update the specified tenants for a collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Arguments:
            `tenants`
                A tenant name, `wvc.config.tenants.Tenant` object, or a list of tenants names
                and/or `wvc.config.tenants.Tenant` objects to update for the given collection.

        Raises:
            `weaviate.WeaviateConnectionError`
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeError`
                If Weaviate reports a non-OK status.
            `weaviate.WeaviateInvalidInputError`
                If `tenants` is not a list of `wvc.Tenant` objects.
        """
        if self.__validate_arguments:
            _validate_input(
                _ValidateArgument(expected=[Tenant, List[Tenant]], name="tenants", value=tenants)
            )

        loaded_tenants = (
            [tenants.model_dump()]
            if isinstance(tenants, Tenant)
            else [tenant.model_dump() for tenant in tenants]
        )

        path = "/schema/" + self.__name + "/tenants"
        self.__connection.put(
            path=path,
            weaviate_object=loaded_tenants,
            error_msg=f"Collection tenants may not have been updated properly for {self.__name}",
            status_codes=_ExpectedStatusCodes(
                ok_in=200, error=f"Update collection tenants for {self.__name}"
            ),
        )

    def exists(self, tenant: Union[str, Tenant]) -> bool:
        """Check if a tenant exists for a collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Arguments:
            `tenant`
                Tenant name or `wvc.config.tenants.Tenant` object to check for existence.

        Returns:
            `bool`
                `True` if the tenant exists, `False` otherwise.

        Raises:
            `weaviate.WeaviateConnectionError`
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeError`
                If Weaviate reports a non-OK status.
        """
        self.__connection._weaviate_version.check_is_at_least_1_25_0("The 'exists' method")

        tenant_name = tenant.name if isinstance(tenant, Tenant) else tenant

        path = "/schema/" + self.__name + "/tenants/" + tenant_name
        response = self.__connection.head(
            path=path,
            error_msg=f"Could not check if tenant exists for {self.__name}",
            status_codes=_ExpectedStatusCodes(
                ok_in=[200, 404], error=f"Check if tenant exists for {self.__name}"
            ),  # allow 404 to perform bool check on response code
        )
        return response.status_code == 200



import time
from typing import List
import uuid

import h5py  # type: ignore

import weaviate
import weaviate.classes as wvc

from _pytest.fixtures import SubRequest

from weaviate.collections.collection import Collection
from .conftest import get_file_path

# The following code is used to test the performance of the weaviate client.
# Download the datasets at
#  - https://storage.googleapis.com/ann-datasets/ann-benchmarks/sift-128-euclidean.hdf5
#  - https://storage.googleapis.com/ann-datasets/ann-benchmarks/dbpedia-openai-1000k-angular.hdf5
#  and place it in the same folder as this file.
#
# In addition to weavaite, you will need to install the "h5py" package.
#
# To run all tests use `pytest -s test_sift.py` in the profiling directory. If you want to run only one of the tests
# you can use the `-k` flag, for example `pytest -s test_sift.py -k test_sift_v3`

EF_VALUES = [16, 32, 64, 128, 256, 512]
LIMIT = 10


def create_schema(
    client: weaviate.WeaviateClient, name: str, efC: int, m: int, shards: int, distance: str
) -> Collection:
    client.collections.delete(name)
    vi = wvc.config.Configure.VectorIndex.hnsw(
        ef_construction=efC,
        max_connections=m,
        ef=-1,
        distance_metric=wvc.config.VectorDistances(distance),
    )

    return client.collections.create(
        name=name,
        vectorizer_config=wvc.config.Configure.Vectorizer.none(),
        vector_index_config=vi,
        properties=[
            wvc.config.Property(
                name="i",
                data_type=wvc.config.DataType.INT,
            )
        ],
        inverted_index_config=wvc.config.Configure.inverted_index(index_timestamps=False),
        sharding_config=wvc.config.Configure.sharding(desired_count=shards),
    )


def load_records_v4(collection: Collection, vectors: List[List[float]]) -> None:
    start = time.time()
    with collection.batch.fixed_size(1000) as batch:
        for i, vector in enumerate(vectors):
            data_object = {"i": i}

            batch.add_object(
                properties=data_object,
                vector=vector,
                uuid=uuid.UUID(int=i),
            )

    print(f"V4: Finished writing {len(vectors)} records in {time.time()-start}s")


def load_records_v3(client: weaviate.Client, vectors: List[List[float]], name: str) -> None:
    start = time.time()

    client.batch.configure(batch_size=1000, num_workers=2)

    with client.batch as batch:
        for i, vector in enumerate(vectors):
            data_object = {"i": i}

            batch.add_data_object(
                data_object=data_object,
                vector=vector,
                class_name=name,
                uuid=uuid.UUID(int=i),
            )

    print(f"V3: Finished writing {len(vectors)} records in {time.time()-start}s")


def query_v4(
    collection: Collection, vectors: List[List[float]], neighbours: List[List[int]], ef: int
) -> None:
    collection.config.update(vector_index_config=wvc.config.Reconfigure.VectorIndex.hnsw(ef=ef))
    start = time.time()
    recall = 0.0

    for i, vec in enumerate(vectors):
        objs = collection.query.near_vector(
            near_vector=list(vec), limit=LIMIT, return_properties=[]
        ).objects
        ideal_neighbors = set(neighbours[i][:LIMIT])
        res_ids = [obj.uuid.int for obj in objs]
        recall += len(ideal_neighbors.intersection(res_ids)) / LIMIT

    print(
        f"V4: Querying {len(vectors)} records with ef {ef} in {time.time()-start}s with recall {recall/len(vectors)}"
    )


def query_v3(
    collection: Collection,
    client: weaviate.Client,
    vectors: List[List[float]],
    neighbours: List[List[int]],
    ef: int,
) -> None:
    collection.config.update(vector_index_config=wvc.config.Reconfigure.VectorIndex.hnsw(ef=ef))
    start = time.time()
    recall = 0.0

    for i, vec in enumerate(vectors):
        res = (
            client.query.get(collection.name, ["i _additional{id}"])
            .with_near_vector(
                {
                    "vector": vec,
                }
            )
            .with_limit(LIMIT)
            .do()
        )
        res_ids = [
            uuid.UUID(res["_additional"]["id"]).int for res in res["data"]["Get"][collection.name]
        ]
        ideal_neighbors = set(neighbours[i][:LIMIT])

        recall += len(ideal_neighbors.intersection(res_ids)) / LIMIT

    print(
        f"V3: Querying {len(vectors)} records with ef {ef} in {time.time()-start}s with recall {recall/len(vectors)}"
    )


def run_v4(file: str, name: str, efc: int, m: int) -> None:
    sift_file = get_file_path(file)

    f = h5py.File(sift_file)
    vectors_import = f["train"]
    vectors_test = f["test"]
    ideal_neighbors = f["neighbors"]

    client = weaviate.connect_to_local()
    collection = create_schema(client, name, efc, m, 1, "l2-squared")
    load_records_v4(collection, vectors_import)
    for ef in EF_VALUES:
        query_v4(collection, vectors_test, ideal_neighbors, ef)


def run_v3(file: str, name: str, efc: int, m: int) -> None:
    sift_file = get_file_path(file)

    f = h5py.File(sift_file)
    vectors_import = f["train"]
    vectors_test = f["test"]
    ideal_neighbors = f["neighbors"]

    client = weaviate.Client(url="http://localhost:8080")

    # use v4 client to create schema to avoid duplicate code
    clientv4 = weaviate.connect_to_local()
    collection = create_schema(clientv4, name, efc, m, 1, "l2-squared")
    load_records_v3(client, vectors_import, name)
    for ef in EF_VALUES:
        query_v3(collection, client, vectors_test, ideal_neighbors, ef)


def test_sift_v3(request: SubRequest) -> None:
    run_v3(file="sift-128-euclidean.hdf5", name=request.node.name, efc=128, m=32)


def test_dbpedia_v3(request: SubRequest) -> None:
    run_v3(file="dbpedia-openai-1000k-angular.hdf5", name=request.node.name, efc=384, m=20)


def test_sift_v4(request: SubRequest) -> None:
    run_v4(file="sift-128-euclidean.hdf5", name=request.node.name, efc=128, m=32)


def test_dbpedia_v4(request: SubRequest) -> None:
    run_v4(file="dbpedia-openai-1000k-angular.hdf5", name=request.node.name, efc=384, m=20)



"""
GraphQL module used to create `get` and/or `aggregate`  GraphQL requests from Weaviate.
"""

__all__ = ["AdditionalProperties", "LinkTo", "Query"]

from .get import AdditionalProperties, LinkTo
from .query import Query



import uuid
from typing import Dict, TypedDict

import pytest as pytest
from typing_extensions import Annotated

from integration.conftest import CollectionFactory, CollectionFactoryGet
from weaviate.collections.classes.config import (
    Configure,
    Property,
    DataType,
    ReferenceProperty,
)
from weaviate.collections.classes.data import DataObject, DataReference
from weaviate.collections.classes.grpc import (
    MetadataQuery,
    QueryReference,
)
from weaviate.collections.classes.internal import (
    CrossReference,
    ReferenceToMulti,
    CrossReferenceAnnotation,
    ReferenceInput,
    SingleReferenceInput,
)
from weaviate.exceptions import WeaviateInvalidInputError
from weaviate.types import UUID

TO_UUID = uuid.UUID("8ad0d33c-8db1-4437-87f3-72161ca2a51a")
TO_UUID2 = uuid.UUID("577887c1-4c6b-5594-aa62-f0c17883d9cf")


@pytest.mark.parametrize("add", [TO_UUID, str(TO_UUID)])
@pytest.mark.parametrize("delete", [TO_UUID, str(TO_UUID)])
def test_reference_add_delete_replace(
    collection_factory: CollectionFactory,
    add: SingleReferenceInput,
    delete: SingleReferenceInput,
) -> None:
    ref_collection = collection_factory(
        name="Target", vectorizer_config=Configure.Vectorizer.none()
    )
    ref_collection.data.insert(properties={}, uuid=TO_UUID)
    collection = collection_factory(
        references=[
            ReferenceProperty.MultiTarget(name="ref", target_collections=[ref_collection.name])
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    uuid_from1 = collection.data.insert({}, uuid=uuid.uuid4())
    uuid_from2 = collection.data.insert({}, references={"ref": TO_UUID}, uuid=uuid.uuid4())
    collection.data.reference_add(from_uuid=uuid_from1, from_property="ref", to=add)

    collection.data.reference_delete(from_uuid=uuid_from1, from_property="ref", to=delete)
    assert (
        len(
            collection.query.fetch_object_by_id(
                uuid_from1, return_references=QueryReference(link_on="ref")
            )
            .references["ref"]
            .objects
        )
        == 0
    )

    collection.data.reference_add(from_uuid=uuid_from2, from_property="ref", to=add)
    obj = collection.query.fetch_object_by_id(
        uuid_from2, return_references=QueryReference(link_on="ref")
    )
    assert obj is not None
    assert len(obj.references["ref"].objects) == 2
    assert TO_UUID in [x.uuid for x in obj.references["ref"].objects]

    collection.data.reference_replace(from_uuid=uuid_from2, from_property="ref", to=[])
    assert (
        len(
            collection.query.fetch_object_by_id(
                uuid_from2, return_references=QueryReference(link_on="ref")
            )
            .references["ref"]
            .objects
        )
        == 0
    )


def test_reference_add_delete_replace_multi_target(
    collection_factory: CollectionFactory,
) -> None:
    ref_collection = collection_factory(
        name="Target", vectorizer_config=Configure.Vectorizer.none()
    )
    ref_collection.data.insert(properties={}, uuid=TO_UUID)
    collection = collection_factory(
        references=[
            ReferenceProperty.MultiTarget(name="ref", target_collections=[ref_collection.name])
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    uuid_from1 = collection.data.insert({}, uuid=uuid.uuid4())
    uuid_from2 = collection.data.insert({}, references={"ref": TO_UUID}, uuid=uuid.uuid4())
    collection.data.reference_add(
        from_uuid=uuid_from1,
        from_property="ref",
        to=ReferenceToMulti(target_collection=ref_collection.name, uuids=TO_UUID),
    )

    collection.data.reference_delete(
        from_uuid=uuid_from1,
        from_property="ref",
        to=ReferenceToMulti(target_collection=ref_collection.name, uuids=TO_UUID),
    )
    assert (
        len(
            collection.query.fetch_object_by_id(
                uuid_from1, return_references=QueryReference(link_on="ref")
            )
            .references["ref"]
            .objects
        )
        == 0
    )

    collection.data.reference_add(
        from_uuid=uuid_from2,
        from_property="ref",
        to=ReferenceToMulti(target_collection=ref_collection.name, uuids=TO_UUID),
    )
    obj = collection.query.fetch_object_by_id(
        uuid_from2, return_references=QueryReference(link_on="ref")
    )
    assert obj is not None
    assert len(obj.references["ref"].objects) == 2
    assert TO_UUID in [x.uuid for x in obj.references["ref"].objects]

    collection.data.reference_replace(
        from_uuid=uuid_from2,
        from_property="ref",
        to=ReferenceToMulti(target_collection=ref_collection.name, uuids=[]),
    )
    assert (
        len(
            collection.query.fetch_object_by_id(
                uuid_from2, return_references=QueryReference(link_on="ref")
            )
            .references["ref"]
            .objects
        )
        == 0
    )


@pytest.mark.parametrize(
    "to",
    [
        [TO_UUID, TO_UUID],
        ReferenceToMulti(target_collection="Target", uuids=[TO_UUID, TO_UUID]),
    ],
)
def test_reference_add_multiple_uuids_error(
    collection_factory: CollectionFactory, to: SingleReferenceInput
) -> None:
    ref_collection = collection_factory(
        name="Target", vectorizer_config=Configure.Vectorizer.none()
    )
    ref_collection.data.insert(properties={}, uuid=TO_UUID)
    collection = collection_factory(
        references=[ReferenceProperty(name="ref", target_collection=ref_collection.name)],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    uuid_from1 = collection.data.insert({}, uuid=uuid.uuid4())
    with pytest.raises(WeaviateInvalidInputError):
        collection.data.reference_add(from_uuid=uuid_from1, from_property="ref", to=to)


@pytest.mark.parametrize(
    "to",
    [
        [TO_UUID, TO_UUID],
        ReferenceToMulti(target_collection="Target", uuids=[TO_UUID, TO_UUID]),
    ],
)
def test_reference_delete_multiple_uuids_error(
    collection_factory: CollectionFactory, to: SingleReferenceInput
) -> None:
    ref_collection = collection_factory(
        name="Target", vectorizer_config=Configure.Vectorizer.none()
    )
    ref_collection.data.insert(properties={}, uuid=TO_UUID)
    collection = collection_factory(
        references=[ReferenceProperty(name="ref", target_collection=ref_collection.name)],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    uuid_from1 = collection.data.insert({}, uuid=uuid.uuid4())
    with pytest.raises(WeaviateInvalidInputError):
        collection.data.reference_delete(from_uuid=uuid_from1, from_property="ref", to=to)


def test_mono_references_grpc(collection_factory: CollectionFactory) -> None:
    A = collection_factory(
        name="A",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
        ],
    )
    uuid_A1 = A.data.insert(properties={"Name": "A1"})
    uuid_A2 = A.data.insert(properties={"Name": "A2"})

    a_objs = A.query.bm25(query="A1", return_properties="name").objects
    assert a_objs[0].collection == A.name
    assert a_objs[0].properties["name"] == "A1"

    B = collection_factory(
        name="B",
        properties=[Property(name="Name", data_type=DataType.TEXT)],
        references=[
            ReferenceProperty(name="a", target_collection=A.name),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    uuid_B = B.data.insert({"Name": "B"}, references={"a": uuid_A1})
    B.data.reference_add(from_uuid=uuid_B, from_property="a", to=uuid_A2)

    b_objs = B.query.bm25(
        query="B",
        return_references=QueryReference(
            link_on="a",
            return_properties=["name"],
        ),
    ).objects
    assert b_objs[0].references["a"].objects[0].collection == A.name
    assert b_objs[0].references["a"].objects[0].properties["name"] == "A1"
    assert b_objs[0].references["a"].objects[0].uuid == uuid_A1
    assert b_objs[0].references["a"].objects[1].collection == A.name
    assert b_objs[0].references["a"].objects[1].properties["name"] == "A2"
    assert b_objs[0].references["a"].objects[1].uuid == uuid_A2

    C = collection_factory(
        name="C",
        properties=[Property(name="Name", data_type=DataType.TEXT)],
        references=[
            ReferenceProperty(name="b", target_collection=B.name),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    C.data.insert({"Name": "find me"}, references={"b": uuid_B})

    c_objs = C.query.bm25(
        query="find",
        return_properties="name",
        return_references=QueryReference(
            link_on="b",
            return_properties="name",
            return_metadata=MetadataQuery(last_update_time=True),
            return_references=QueryReference(
                link_on="a",
                return_properties="name",
            ),
        ),
    ).objects
    assert c_objs[0].collection == C.name
    assert c_objs[0].properties["name"] == "find me"
    assert c_objs[0].references["b"].objects[0].collection == B.name
    assert c_objs[0].references["b"].objects[0].properties["name"] == "B"
    assert c_objs[0].references["b"].objects[0].metadata.last_update_time is not None
    assert c_objs[0].references["b"].objects[0].references["a"].objects[0].collection == A.name
    assert (
        c_objs[0].references["b"].objects[0].references["a"].objects[0].properties["name"] == "A1"
    )
    assert c_objs[0].references["b"].objects[0].references["a"].objects[1].collection == A.name
    assert (
        c_objs[0].references["b"].objects[0].references["a"].objects[1].properties["name"] == "A2"
    )


@pytest.mark.parametrize("level", ["col-col", "col-query", "query-col", "query-query"])
def test_mono_references_grpc_with_generics(
    collection_factory: CollectionFactory,
    collection_factory_get: CollectionFactoryGet,
    level: str,
) -> None:
    class AProps(TypedDict):
        name: str

    class BProps(TypedDict):
        name: str

    class BRefs(TypedDict):
        a: Annotated[
            CrossReference[AProps, None],
            CrossReferenceAnnotation(
                metadata=MetadataQuery(creation_time=True), include_vector=True
            ),
        ]

    class CProps(TypedDict):
        name: str

    class CRefs(TypedDict):
        b: CrossReference[BProps, BRefs]

    dummy_a = collection_factory(
        name="a",
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(
            vectorize_collection_name=False
        ),
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
        ],
    )
    A = collection_factory_get(dummy_a.name, AProps)
    uuid_A1 = A.data.insert(AProps(name="A1"))
    uuid_A2 = A.data.insert(AProps(name="A2"))

    dummy_b = collection_factory(
        name="B",
        properties=[Property(name="Name", data_type=DataType.TEXT)],
        references=[
            ReferenceProperty(name="a", target_collection=A.name),
        ],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(
            vectorize_collection_name=False
        ),
    )
    B = collection_factory_get(dummy_b.name, BProps)
    uuid_B = B.data.insert(properties={"name": "B"}, references={"a": uuid_A1})
    B.data.reference_add(
        from_uuid=uuid_B,
        from_property="a",
        to=uuid_A2,
    )

    b_objs = B.query.bm25(query="B", return_references=BRefs).objects
    assert b_objs[0].collection == B.name
    assert b_objs[0].properties["name"] == "B"
    assert b_objs[0].references["a"].objects[0].collection == A.name
    assert b_objs[0].references["a"].objects[0].properties["name"] == "A1"
    assert b_objs[0].references["a"].objects[0].uuid == uuid_A1
    assert b_objs[0].references["a"].objects[0].references is None
    assert b_objs[0].references["a"].objects[1].collection == A.name
    assert b_objs[0].references["a"].objects[1].properties["name"] == "A2"
    assert b_objs[0].references["a"].objects[1].uuid == uuid_A2
    assert b_objs[0].references["a"].objects[1].references is None

    dummy_c = collection_factory(
        name="C",
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
            Property(name="Age", data_type=DataType.INT),
        ],
        references=[
            ReferenceProperty(name="b", target_collection=B.name),
        ],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(
            vectorize_collection_name=False
        ),
    )
    C = collection_factory_get(dummy_c.name, CProps)
    C.data.insert(properties={"name": "find me"}, references={"b": uuid_B})

    if level == "col-col":
        c_objs = (
            collection_factory_get(C.name, CProps, CRefs)
            .query.bm25(query="find", include_vector=True)
            .objects
        )
    elif level == "col-query":
        c_objs = (
            collection_factory_get(C.name, CProps)
            .query.bm25(
                query="find",
                include_vector=True,
                return_references=CRefs,
            )
            .objects
        )
    elif level == "query-col":
        c_objs = (
            collection_factory_get(C.name, data_model_refs=CRefs)
            .query.bm25(
                query="find",
                include_vector=True,
                return_properties=CProps,
            )
            .objects
        )
    else:
        c_objs = (
            collection_factory_get(C.name)
            .query.bm25(
                query="find",
                include_vector=True,
                return_properties=CProps,
                return_references=CRefs,
            )
            .objects
        )
    assert c_objs[0].collection == C.name
    assert (
        c_objs[0].properties["name"] == "find me"
    )  # happy path (in type and in return_properties)
    assert c_objs[0].uuid is not None
    assert "default" in c_objs[0].vector
    assert (
        c_objs[0].properties.get("not_specified") is None
    )  # type is str but instance is None (in type but not in return_properties)
    assert c_objs[0].references["b"].objects[0].collection == B.name
    assert c_objs[0].references["b"].objects[0].properties["name"] == "B"
    assert c_objs[0].references["b"].objects[0].uuid == uuid_B
    assert "default" not in c_objs[0].references["b"].objects[0].vector
    assert c_objs[0].references["b"].objects[0].references["a"].objects[0].collection == A.name
    assert (
        c_objs[0].references["b"].objects[0].references["a"].objects[0].properties["name"] == "A1"
    )
    assert c_objs[0].references["b"].objects[0].references["a"].objects[0].uuid == uuid_A1
    assert (
        c_objs[0].references["b"].objects[0].references["a"].objects[0].metadata.creation_time
        is not None
    )
    assert "default" in c_objs[0].references["b"].objects[0].references["a"].objects[0].vector
    assert c_objs[0].references["b"].objects[0].references["a"].objects[1].collection == A.name
    assert (
        c_objs[0].references["b"].objects[0].references["a"].objects[1].properties["name"] == "A2"
    )
    assert c_objs[0].references["b"].objects[0].references["a"].objects[1].uuid == uuid_A2
    assert (
        c_objs[0].references["b"].objects[0].references["a"].objects[1].metadata.creation_time
        is not None
    )
    assert "default" in c_objs[0].references["b"].objects[0].references["a"].objects[1].vector


def test_multi_references_grpc(collection_factory: CollectionFactory) -> None:
    A = collection_factory(
        name="A",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
        ],
    )
    uuid_A = A.data.insert(properties={"Name": "A"})

    B = collection_factory(
        name="B",
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    uuid_B = B.data.insert({"Name": "B"})

    C = collection_factory(
        name="C",
        properties=[Property(name="Name", data_type=DataType.TEXT)],
        references=[
            ReferenceProperty.MultiTarget(name="ref", target_collections=[A.name, B.name]),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    C.data.insert(
        {
            "Name": "first",
        },
        references={
            "ref": ReferenceToMulti(uuids=uuid_A, target_collection=A.name),
        },
    )
    C.data.insert(
        {
            "Name": "second",
        },
        references={
            "ref": ReferenceToMulti(uuids=uuid_B, target_collection=B.name),
        },
    )

    objects = C.query.bm25(
        query="first",
        return_properties="name",
        return_references=QueryReference.MultiTarget(
            link_on="ref",
            target_collection=A.name,
            return_properties=["name"],
            return_metadata=MetadataQuery(last_update_time=True),
        ),
    ).objects
    assert objects[0].collection == C.name
    assert objects[0].properties["name"] == "first"
    assert len(objects[0].references["ref"].objects) == 1
    assert objects[0].references["ref"].objects[0].collection == A.name
    assert objects[0].references["ref"].objects[0].properties["name"] == "A"
    assert objects[0].references["ref"].objects[0].metadata.last_update_time is not None

    objects = C.query.bm25(
        query="second",
        return_properties="name",
        return_references=QueryReference.MultiTarget(
            link_on="ref",
            target_collection=B.name,
            return_properties=[
                "name",
            ],
            return_metadata=MetadataQuery(last_update_time=True),
        ),
    ).objects
    assert objects[0].collection == C.name
    assert objects[0].properties["name"] == "second"
    assert len(objects[0].references["ref"].objects) == 1
    assert objects[0].references["ref"].objects[0].collection == B.name
    assert objects[0].references["ref"].objects[0].properties["name"] == "B"
    assert objects[0].references["ref"].objects[0].metadata.last_update_time is not None


def test_multi_references_grpc_with_generics(collection_factory: CollectionFactory) -> None:
    A = collection_factory(
        name="A",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
        ],
    )
    uuid_A = A.data.insert(properties={"Name": "A"})

    B = collection_factory(
        name="B",
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    uuid_B = B.data.insert({"Name": "B"})

    C = collection_factory(
        name="C",
        properties=[Property(name="Name", data_type=DataType.TEXT)],
        references=[
            ReferenceProperty.MultiTarget(name="ref", target_collections=[A.name, B.name]),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    C.data.insert(
        {
            "Name": "first",
        },
        references={
            "ref": ReferenceToMulti(uuids=uuid_A, target_collection=A.name),
        },
    )
    C.data.insert(
        {
            "Name": "second",
        },
        references={
            "ref": ReferenceToMulti(uuids=uuid_B, target_collection=B.name),
        },
    )

    class AProps(TypedDict):
        name: str

    class BProps(TypedDict):
        name: str

    class CProps(TypedDict):
        name: str

    class CRefsA(TypedDict):
        ref: Annotated[
            CrossReference[AProps, None],
            CrossReferenceAnnotation(
                metadata=MetadataQuery(last_update_time=True), target_collection=A.name
            ),
        ]

    class CRefsB(TypedDict):
        ref: Annotated[
            CrossReference[BProps, None],
            CrossReferenceAnnotation(
                metadata=MetadataQuery(last_update_time=True), target_collection=B.name
            ),
        ]

    objects = C.query.bm25(
        query="first",
        return_properties=CProps,
        return_references=CRefsA,
    ).objects
    assert objects[0].collection == C.name
    assert objects[0].properties["name"] == "first"
    assert len(objects[0].references["ref"].objects) == 1
    assert objects[0].references["ref"].objects[0].collection == A.name
    assert objects[0].references["ref"].objects[0].properties["name"] == "A"
    assert objects[0].references["ref"].objects[0].metadata.last_update_time is not None

    objects = C.query.bm25(
        query="second",
        return_properties=CProps,
        return_references=CRefsB,
    ).objects
    assert objects[0].collection == C.name
    assert objects[0].properties["name"] == "second"
    assert len(objects[0].references["ref"].objects) == 1
    assert objects[0].references["ref"].objects[0].collection == B.name
    assert objects[0].references["ref"].objects[0].properties["name"] == "B"
    assert objects[0].references["ref"].objects[0].metadata.last_update_time is not None


def test_references_batch(collection_factory: CollectionFactory) -> None:
    ref_collection = collection_factory(
        name="To",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[Property(name="number", data_type=DataType.INT)],
    )
    num_objects = 10

    uuids_to = list(
        ref_collection.data.insert_many(
            [DataObject(properties={"number": i}) for i in range(num_objects)]
        ).uuids.values()
    )
    collection = collection_factory(
        name="From",
        properties=[
            Property(name="num", data_type=DataType.INT),
        ],
        references=[ReferenceProperty(name="ref", target_collection=ref_collection.name)],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    uuids_from = list(
        collection.data.insert_many(
            [DataObject(properties={"num": i}) for i in range(num_objects)]
        ).uuids.values()
    )

    # use both ways of adding batch-references in two calls to preserve order:
    # - single points to an object with the same value as property
    # - multi always points to the first 3 objects
    batch_return = collection.data.reference_add_many(
        [
            DataReference(
                from_property="ref",
                from_uuid=uuids_from[i],
                to_uuid=uuids_to[i],
            )
            for i in range(num_objects)
        ]
    )
    assert batch_return.has_errors is False

    batch_return = collection.data.reference_add_many(
        [
            DataReference(
                from_property="ref",
                from_uuid=uuids_from[i],
                to_uuid=[uuids_to[j] for j in range(3)],
            )
            for i in range(num_objects)
        ]
    )
    assert batch_return.has_errors is False

    objects = collection.query.fetch_objects(
        return_properties=["num"],
        return_references=[QueryReference(link_on="ref")],
    ).objects

    for obj in objects:
        assert obj.properties["num"] == obj.references["ref"].objects[0].properties["number"]
        assert obj.references["ref"].objects[0].uuid in uuids_to
        assert len(obj.references["ref"].objects) == 4
        refs = [obj.references["ref"].objects[j + 1].properties["number"] for j in range(3)]
        refs.sort()
        assert [0, 1, 2] == refs


def test_batch_reference_multi_taret(collection_factory: CollectionFactory) -> None:
    to_collection = collection_factory(
        name="To",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[Property(name="number", data_type=DataType.INT)],
    )
    from_collection = collection_factory(
        name="From",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[Property(name="num", data_type=DataType.INT)],
    )
    from_collection.config.add_reference(
        ReferenceProperty.MultiTarget(
            name="ref", target_collections=[to_collection.name, from_collection.name]
        )
    )

    num_objects = 5

    uuids_to = list(
        to_collection.data.insert_many(
            [DataObject(properties={"number": i}) for i in range(num_objects)]
        ).uuids.values()
    )
    uuids_from = list(
        from_collection.data.insert_many(
            [DataObject(properties={"num": i}) for i in range(num_objects)]
        ).uuids.values()
    )

    # add to to_collection
    batch_return = from_collection.data.reference_add_many(
        [
            DataReference.MultiTarget(
                from_property="ref",
                from_uuid=uuids_from[i],
                to_uuid=uuids_to[i],
                target_collection=to_collection.name,
            )
            for i in range(num_objects)
        ]
    )
    assert batch_return.has_errors is False

    # add to from_collection
    batch_return = from_collection.data.reference_add_many(
        [
            DataReference.MultiTarget(
                from_property="ref",
                from_uuid=uuids_from[i],
                to_uuid=uuids_from[i],
                target_collection=from_collection.name,
            )
            for i in range(num_objects)
        ]
    )
    assert batch_return.has_errors is False

    objects_with_to_ref = from_collection.query.fetch_objects(
        return_properties=["num"],
        return_references=[
            QueryReference.MultiTarget(link_on="ref", target_collection=to_collection.name)
        ],
    ).objects
    for obj in objects_with_to_ref:
        assert obj.properties["num"] == obj.references["ref"].objects[0].properties["number"]

    objects_with_from_ref = from_collection.query.fetch_objects(
        return_properties=["num"],
        return_references=[
            QueryReference.MultiTarget(link_on="ref", target_collection=from_collection.name)
        ],
    ).objects
    for obj in objects_with_from_ref:
        assert obj.properties["num"] == obj.references["ref"].objects[0].properties["num"]


def test_insert_many_with_refs(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[Property(name="Name", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    collection.config.add_reference(
        ReferenceProperty(name="self", target_collection=collection.name)
    )

    uuid1 = collection.data.insert({"name": "A"})
    uuid2 = collection.data.insert({"name": "B"})

    batch_return = collection.data.insert_many(
        [
            DataObject(
                properties={"name": "C"},
                references={"self": uuid1},
            ),
            DataObject(
                properties={"name": "D"},
                references={"self": [uuid1, uuid2]},
            ),
            DataObject(
                properties={"name": "E"},
                references={"self": uuid1},
            ),
            DataObject(
                properties={"name": "F"},
                references={"self": [uuid1, uuid2]},
            ),
        ]
    )
    assert batch_return.has_errors is False

    for obj in collection.query.fetch_objects(
        return_properties=["name"], return_references=QueryReference(link_on="self")
    ).objects:
        if obj.properties["name"] in ["A", "B"]:
            assert obj.references == {}
        else:
            assert obj.references is not None
            if obj.properties["name"] == "C":
                assert obj.references["self"].objects[0].uuid == uuid1
            elif obj.properties["name"] == "D":
                assert obj.references["self"].objects[0].uuid == uuid1
                assert obj.references["self"].objects[1].uuid == uuid2
            elif obj.properties["name"] == "E":
                assert obj.references["self"].objects[0].uuid == uuid1
            elif obj.properties["name"] == "F":
                assert obj.references["self"].objects[0].uuid == uuid1
                assert obj.references["self"].objects[1].uuid == uuid2


def test_references_batch_with_errors(collection_factory: CollectionFactory) -> None:
    to = collection_factory(
        name="To",
        vectorizer_config=Configure.Vectorizer.none(),
    )

    collection = collection_factory(
        name="From",
        properties=[
            Property(name="num", data_type=DataType.INT),
        ],
        references=[ReferenceProperty(name="ref", target_collection=to.name)],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    batch_return = collection.data.reference_add_many(
        [DataReference(from_property="doesNotExist", from_uuid=uuid.uuid4(), to_uuid=uuid.uuid4())],
    )
    assert batch_return.has_errors is True
    assert 0 in batch_return.errors


# commented out due to mypy failures since it is stale code
# @pytest.mark.skip(reason="string syntax has been temporarily removed from the API")
# def test_references_with_string_syntax(client: weaviate.WeaviateClient):
#     name1 = "TestReferencesWithStringSyntaxA"
#     name2 = "TestReferencesWithStringSyntaxB"
#     client.collections.delete(name1)
#     client.collections.delete(name2)

#     client.collections.create(
#         name=name1,
#         vectorizer_config=Configure.Vectorizer.none(),
#         properties=[
#             Property(name="Name", data_type=DataType.TEXT),
#             Property(name="Age", data_type=DataType.INT),
#             Property(name="Weird__Name", data_type=DataType.INT),
#         ],
#     )

#     uuid_A = client.collections.get(name1).data.insert(
#         properties={"Name": "A", "Age": 1, "Weird__Name": 2}
#     )

#     client.collections.get(name1).query.fetch_object_by_id(uuid_A)

#     client.collections.create(
#         name=name2,
#         properties=[
#             Property(name="Name", data_type=DataType.TEXT),
#         ],
#         references=[ReferenceProperty(name="ref", target_collection=name1)],
#         vectorizer_config=Configure.Vectorizer.none(),
#     )

#     client.collections.get(name2).data.insert(
#         {"Name": "B"}, references={"ref": reference_to_no_warning(uuids=uuid_A)}
#     )

#     objects = (
#         client.collections.get(name2)
#         .query.bm25(
#             query="B",
#             return_properties=[
#                 "name",
#                 "__ref__properties__Name",
#                 "__ref__properties__Age",
#                 "__ref__properties__Weird__Name",
#                 "__ref__metadata__last_update_time_unix",
#             ],
#         )
#         .objects
#     )

#     assert objects[0].properties["name"] == "B"
#     assert objects[0].references["ref"].objects[0].properties["name"] == "A"
#     assert objects[0].references["ref"].objects[0].properties["age"] == 1
#     assert objects[0].references["ref"].objects[0].properties["weird__Name"] == 2
#     assert objects[0].references["ref"].objects[0].uuid == uuid_A
#     assert objects[0].references["ref"].objects[0].metadata.last_update_time_unix is not None


def test_object_without_references(collection_factory: CollectionFactory) -> None:
    to = collection_factory(name="To", vectorizer_config=Configure.Vectorizer.none())

    source = collection_factory(
        name="From",
        references=[
            ReferenceProperty(name="ref_partial", target_collection=to.name),
            ReferenceProperty(name="ref_full", target_collection=to.name),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    uuid_to = to.data.insert(properties={})

    uuid_from1 = source.data.insert(
        references={"ref_partial": uuid_to, "ref_full": uuid_to},
        properties={},
    )
    uuid_from2 = source.data.insert(references={"ref_full": uuid_to}, properties={})

    obj1 = source.query.fetch_object_by_id(
        uuid_from2,
        return_references=[
            QueryReference(link_on="ref_full"),
            QueryReference(link_on="ref_partial"),
        ],
    )
    assert "ref_full" in obj1.references and "ref_partial" not in obj1.references
    assert obj1.collection == source.name

    obj2 = source.query.fetch_object_by_id(
        uuid_from1,
        return_references=[
            QueryReference(link_on="ref_full"),
            QueryReference(link_on="ref_partial"),
        ],
    )
    assert "ref_full" in obj2.references and "ref_partial" in obj2.references
    assert obj2.collection == source.name


def test_ref_case_sensitivity(collection_factory: CollectionFactory) -> None:
    to = collection_factory(name="To", vectorizer_config=Configure.Vectorizer.none())

    source = collection_factory(
        name="From",
        references=[
            ReferenceProperty(name="ref", target_collection=to.name),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    # added as upper-case UUID
    uuid_upper_str = "4E5CD755-4F43-44C5-B23C-0C7D6F6C21E6"
    to.data.insert(uuid=uuid_upper_str, properties={})

    # adding a ref as lower-case UUID should work
    from1 = source.data.insert(properties={}, references={"ref": uuid_upper_str.lower()})

    # try to add as upper-case UUID via different methods
    from2 = source.data.insert(properties={}, references={"ref": uuid_upper_str})
    from3 = source.data.insert(properties={})
    source.data.reference_add(from_uuid=from3, from_property="ref", to=uuid_upper_str)

    from4 = source.data.insert(properties={})
    source.data.reference_add_many(
        [DataReference(from_uuid=from4, from_property="ref", to_uuid=uuid_upper_str)]
    )

    from5 = source.data.insert_many(
        [DataObject(properties={}, references={"ref": uuid_upper_str})]
    ).uuids[0]

    for uid in [from1, from2, from3, from4, from5]:
        obj = source.query.fetch_object_by_id(
            uid, return_references=[QueryReference(link_on="ref")]
        )
        assert "ref" in obj.references


def test_empty_return_reference(collection_factory: CollectionFactory) -> None:
    to = collection_factory(name="To", vectorizer_config=Configure.Vectorizer.none())
    source = collection_factory(
        name="From",
        references=[
            ReferenceProperty(name="ref", target_collection=to.name),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    uuid_source = source.data.insert(properties={})
    obj = source.query.fetch_object_by_id(
        uuid_source, return_references=[QueryReference(link_on="ref")]
    )
    assert obj.references == {}


@pytest.mark.parametrize(
    "to_uuid",
    [TO_UUID, str(TO_UUID), [TO_UUID], [str(TO_UUID)]],
)
def test_refs_different_input_insert(
    collection_factory: CollectionFactory, to_uuid: ReferenceInput
) -> None:
    to = collection_factory(name="To", vectorizer_config=Configure.Vectorizer.none())
    to.data.insert(properties={}, uuid=TO_UUID)

    source = collection_factory(
        name="From",
        references=[
            ReferenceProperty(name="ref", target_collection=to.name),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    from_uuid = source.data.insert(properties={}, references={"ref": to_uuid})
    obj = source.query.fetch_object_by_id(
        from_uuid, return_references=[QueryReference(link_on="ref")]
    )
    assert obj.references["ref"].objects[0].uuid == TO_UUID


@pytest.mark.parametrize(
    "to_uuid",
    [TO_UUID, str(TO_UUID), [TO_UUID], [str(TO_UUID)]],
)
def test_refs_different_input_insert_many(
    collection_factory: CollectionFactory, to_uuid: ReferenceInput
) -> None:
    to = collection_factory(name="To", vectorizer_config=Configure.Vectorizer.none())
    to.data.insert(properties={}, uuid=TO_UUID)

    source = collection_factory(
        name="From",
        references=[
            ReferenceProperty(name="ref", target_collection=to.name),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    source.config.add_reference(
        ReferenceProperty.MultiTarget(name="multi", target_collections=[to.name, source.name])
    )

    from_uuid = source.data.insert_many([DataObject(properties={}, references={"ref": to_uuid})])
    assert not from_uuid.has_errors
    assert len(from_uuid.uuids) == 1
    obj = source.query.fetch_object_by_id(
        from_uuid.uuids[0], return_references=[QueryReference(link_on="ref")]
    )
    assert obj.references["ref"].objects[0].uuid == TO_UUID

    from_uuid4 = source.data.insert_many(
        [
            DataObject(
                properties={},
                references={"multi": ReferenceToMulti(uuids=TO_UUID, target_collection=to.name)},
            )
        ]
    )
    assert not from_uuid4.has_errors
    assert len(from_uuid4.uuids) == 1
    obj = source.query.fetch_object_by_id(
        from_uuid4.uuids[0],
        return_references=[QueryReference.MultiTarget(link_on="multi", target_collection=to.name)],
    )
    assert obj.references["multi"].objects[0].uuid == TO_UUID


@pytest.mark.parametrize("to_uuid", [TO_UUID, str(TO_UUID)])
def test_refs_different_reference_add(collection_factory: CollectionFactory, to_uuid: str) -> None:
    to = collection_factory(name="To", vectorizer_config=Configure.Vectorizer.none())
    to.data.insert(properties={}, uuid=TO_UUID)

    source = collection_factory(
        name="From",
        references=[
            ReferenceProperty(name="ref", target_collection=to.name),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    from_uuid = source.data.insert(properties={})

    source.data.reference_add(from_property="ref", from_uuid=from_uuid, to=to_uuid)

    obj = source.query.fetch_object_by_id(
        from_uuid, return_references=[QueryReference(link_on="ref")]
    )
    assert obj.references["ref"].objects[0].uuid == TO_UUID


@pytest.mark.parametrize("to_uuid", [TO_UUID, str(TO_UUID), [TO_UUID], [str(TO_UUID)]])
def test_refs_different_reference_add_many(
    collection_factory: CollectionFactory, to_uuid: UUID
) -> None:
    to = collection_factory(name="To", vectorizer_config=Configure.Vectorizer.none())
    to.data.insert(properties={}, uuid=TO_UUID)
    to.data.insert(properties={}, uuid=TO_UUID2)

    source = collection_factory(
        name="From",
        references=[
            ReferenceProperty(name="ref", target_collection=to.name),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    from_uuid = source.data.insert(properties={})

    source.data.reference_add_many(
        [DataReference(from_property="ref", from_uuid=from_uuid, to_uuid=to_uuid)]
    )

    obj = source.query.fetch_object_by_id(
        from_uuid, return_references=[QueryReference(link_on="ref")]
    )
    assert obj.references["ref"].objects[0].uuid == TO_UUID


@pytest.mark.parametrize(
    "to_uuid",
    [TO_UUID2, str(TO_UUID2), [TO_UUID2], [str(TO_UUID2)]],
)
def test_refs_different_reference_replace(
    collection_factory: CollectionFactory, to_uuid: ReferenceInput
) -> None:
    to = collection_factory(name="To", vectorizer_config=Configure.Vectorizer.none())
    to.data.insert(properties={}, uuid=TO_UUID)
    to.data.insert(properties={}, uuid=TO_UUID2)

    source = collection_factory(
        name="From",
        references=[
            ReferenceProperty(name="ref", target_collection=to.name),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    from_uuid = source.data.insert(properties={})
    source.data.reference_add(from_property="ref", from_uuid=from_uuid, to=TO_UUID)
    obj = source.query.fetch_object_by_id(
        from_uuid, return_references=[QueryReference(link_on="ref")]
    )
    assert obj.references["ref"].objects[0].uuid == TO_UUID

    source.data.reference_replace(from_property="ref", from_uuid=from_uuid, to=to_uuid)
    obj = source.query.fetch_object_by_id(
        from_uuid, return_references=[QueryReference(link_on="ref")]
    )
    assert obj.references["ref"].objects[0].uuid == TO_UUID2


def test_bad_generic_return_references(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[Property(name="Name", data_type=DataType.TEXT)],
    )

    class SomeGeneric(TypedDict):
        field: int

    uuid = collection.data.insert(properties={"Name": "A"})
    with pytest.raises(WeaviateInvalidInputError):
        collection.query.fetch_object_by_id(
            uuid,
            return_references=SomeGeneric,
        )

    class OtherGeneric(TypedDict):
        field: Annotated[
            int,
            CrossReferenceAnnotation(metadata=MetadataQuery(creation_time=True)),
        ]

    with pytest.raises(WeaviateInvalidInputError):
        collection.query.fetch_object_by_id(
            uuid,
            return_references=SomeGeneric,
        )


def test_generic_type_hint_return_references(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[Property(name="Name", data_type=DataType.TEXT)],
        data_model_refs=Dict[str, CrossReference[Dict[str, str], None]],
    )
    collection.config.add_reference(
        ReferenceProperty(name="self", target_collection=collection.name)
    )

    uuid1 = collection.data.insert(properties={"Name": "A"})
    uuid2 = collection.data.insert(properties={"Name": "B"}, references={"self": uuid1})
    obj = collection.query.fetch_object_by_id(
        uuid2,
        return_references=QueryReference(
            link_on="self",
        ),
    )
    assert obj.properties == {"name": "B"}
    assert obj.references["self"].objects[0].uuid == uuid1
    assert obj.references["self"].objects[0].properties == {"name": "A"}



from weaviate.backup.backup import (
    BackupCompressionLevel,
    BackupConfigCreate,
    BackupConfigRestore,
    BackupStorage,
)

__all__ = [
    "BackupCompressionLevel",
    "BackupConfigCreate",
    "BackupConfigRestore",
    "BackupStorage",
]



from typing import TYPE_CHECKING, Generic, List, Optional, Union

from weaviate.collections.batch.base import (
    _BatchBase,
    _BatchDataWrapper,
    _BatchMode,
    _DynamicBatching,
    _FixedSizeBatching,
    _RateLimitedBatching,
)
from weaviate.collections.batch.batch_wrapper import _BatchWrapper, _ContextManagerWrapper
from weaviate.collections.classes.config import ConsistencyLevel, Vectorizers
from weaviate.collections.classes.internal import ReferenceInputs, ReferenceInput
from weaviate.collections.classes.types import Properties
from weaviate.connect import ConnectionV4
from weaviate.types import UUID, VECTORS

if TYPE_CHECKING:
    from weaviate.collections.config import _ConfigCollection


class _BatchCollection(Generic[Properties], _BatchBase):
    def __init__(
        self,
        connection: ConnectionV4,
        consistency_level: Optional[ConsistencyLevel],
        results: _BatchDataWrapper,
        batch_mode: _BatchMode,
        name: str,
        tenant: Optional[str],
        vectorizer_batching: bool,
    ) -> None:
        super().__init__(
            connection=connection,
            consistency_level=consistency_level,
            results=results,
            batch_mode=batch_mode,
            vectorizer_batching=vectorizer_batching,
        )
        self.__name = name
        self.__tenant = tenant

    def add_object(
        self,
        properties: Optional[Properties] = None,
        references: Optional[ReferenceInputs] = None,
        uuid: Optional[UUID] = None,
        vector: Optional[VECTORS] = None,
    ) -> UUID:
        """Add one object to this batch.

        NOTE: If the UUID of one of the objects already exists then the existing object will be replaced by the new object.

        Arguments:
            `properties`
                The data properties of the object to be added as a dictionary.
            `references`
                The references of the object to be added as a dictionary.
            `uuid`:
                The UUID of the object as an uuid.UUID object or str. If it is None an UUIDv4 will generated, by default None
            `vector`:
                The embedding of the object. Can be used when a collection does not have a vectorization module or the given
                vector was generated using the _identical_ vectorization module that is configured for the class. In this
                case this vector takes precedence.
                Supported types are
                - for single vectors: `list`, 'numpy.ndarray`, `torch.Tensor` and `tf.Tensor`, by default None.
                - for named vectors: Dict[str, *list above*], where the string is the name of the vector.

        Returns:
            `str`
                The UUID of the added object. If one was not provided a UUIDv4 will be auto-generated for you and returned here.

        Raises:
            `WeaviateBatchValidationError`
                If the provided options are in the format required by Weaviate.
        """
        return self._add_object(
            collection=self.__name,
            properties=properties,
            references=references,
            uuid=uuid,
            vector=vector,
            tenant=self.__tenant,
        )

    def add_reference(
        self, from_uuid: UUID, from_property: str, to: Union[ReferenceInput, List[UUID]]
    ) -> None:
        """Add a reference to this batch.

        Arguments:
            `from_uuid`
                The UUID of the object, as an uuid.UUID object or str, that should reference another object.
            `from_property`
                The name of the property that contains the reference.
            `to`
                The UUID of the referenced object, as an uuid.UUID object or str, that is actually referenced.
                For multi-target references use wvc.Reference.to_multi_target().

        Raises:
            `WeaviateBatchValidationError`
                If the provided options are in the format required by Weaviate.
        """
        self._add_reference(
            from_uuid,
            self.__name,
            from_property,
            to,
            self.__tenant,
        )


class _BatchCollectionWrapper(Generic[Properties], _BatchWrapper):
    def __init__(
        self,
        connection: ConnectionV4,
        consistency_level: Optional[ConsistencyLevel],
        name: str,
        tenant: Optional[str],
        config: "_ConfigCollection",
    ) -> None:
        super().__init__(connection, consistency_level)
        self.__name = name
        self.__tenant = tenant
        self.__config = config
        self._vectorizer_batching: Optional[bool] = None

    def __create_batch_and_reset(self) -> _ContextManagerWrapper[_BatchCollection[Properties]]:
        if self._vectorizer_batching is None:
            config = self.__config.get(simple=True)
            if config.vector_config is not None:
                vectorizer_batching = False
                for vec_config in config.vector_config.values():
                    if vec_config.vectorizer.vectorizer is not Vectorizers.NONE:
                        vectorizer_batching = True
                        break
                self._vectorizer_batching = vectorizer_batching
            else:
                self._vectorizer_batching = config.vectorizer is not Vectorizers.NONE

        self._batch_data = _BatchDataWrapper()  # clear old data
        return _ContextManagerWrapper(
            _BatchCollection[Properties](
                connection=self._connection,
                consistency_level=self._consistency_level,
                results=self._batch_data,
                batch_mode=self._batch_mode,
                name=self.__name,
                tenant=self.__tenant,
                vectorizer_batching=self._vectorizer_batching,
            )
        )

    def dynamic(self) -> _ContextManagerWrapper[_BatchCollection[Properties]]:
        """Configure dynamic batching.

        When you exit the context manager, the final batch will be sent automatically.
        """
        self._batch_mode: _BatchMode = _DynamicBatching()
        return self.__create_batch_and_reset()

    def fixed_size(
        self, batch_size: int = 100, concurrent_requests: int = 2
    ) -> _ContextManagerWrapper[_BatchCollection[Properties]]:
        """Configure fixed size batches. Note that the default is dynamic batching.

        When you exit the context manager, the final batch will be sent automatically.

        Arguments:
            `batch_size`
                The number of objects/references to be sent in one batch. If not provided, the default value is 100.
            `concurrent_requests`
                The number of concurrent requests when sending batches. This controls the number of concurrent requests
                made to Weaviate and not the speed of batch creation within Python.
        """
        self._batch_mode = _FixedSizeBatching(batch_size, concurrent_requests)
        return self.__create_batch_and_reset()

    def rate_limit(
        self, requests_per_minute: int
    ) -> _ContextManagerWrapper[_BatchCollection[Properties]]:
        """Configure batches with a rate limited vectorizer.

        When you exit the context manager, the final batch will be sent automatically.

        Arguments:
            `requests_per_minute`
                The number of requests that the vectorizer can process per minute.
        """
        self._batch_mode = _RateLimitedBatching(requests_per_minute)
        return self.__create_batch_and_reset()



from weaviate.collections.classes.config import (
    BM25Config,
    CollectionConfig,
    CollectionConfigSimple,
    GenerativeConfig,
    GenerativeSearches,
    InvertedIndexConfig,
    MultiTenancyConfig,
    PQConfig,
    PQEncoderConfig,
    PQEncoderDistribution,
    PQEncoderType,
    PropertyConfig,
    PropertyType,
    ReferencePropertyConfig,
    ReplicationConfig,
    Rerankers,
    RerankerConfig,
    ShardingConfig,
    ShardStatus,
    ShardTypes,
    VectorDistances,
    VectorIndexConfigHNSW,
    VectorIndexConfigFlat,
    VectorIndexType,
    Vectorizers,
    VectorizerConfig,
)

__all__ = [
    "BM25Config",
    "CollectionConfig",
    "CollectionConfigSimple",
    "GenerativeConfig",
    "GenerativeSearches",
    "InvertedIndexConfig",
    "MultiTenancyConfig",
    "PQConfig",
    "PQEncoderConfig",
    "PQEncoderDistribution",
    "PQEncoderType",
    "PropertyConfig",
    "PropertyType",
    "ReferencePropertyConfig",
    "ReplicationConfig",
    "Rerankers",
    "RerankerConfig",
    "ShardingConfig",
    "ShardStatus",
    "ShardTypes",
    "VectorDistances",
    "VectorIndexConfigHNSW",
    "VectorIndexConfigFlat",
    "VectorIndexType",
    "Vectorizers",
    "VectorizerConfig",
]



import weaviate
from weaviate.client import WeaviateClient
from weaviate.collections.classes.config import DataType, Property, Configure


def test_raw_gql_v4() -> None:
    name = "RawGraphQlTest"
    number = 20
    client: WeaviateClient = weaviate.connect_to_local()
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=[Property(name="Name", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    collection.data.insert_many([{"Name": f"name {i}"} for i in range(number)])

    response = client.graphql_raw_query(
        """{
                                        Aggregate {RawGraphQlTest{meta {count}}}
                                        Get{RawGraphQlTest{name}}
                                        }"""
    )

    assert response.errors is None
    assert response.aggregate[name][0]["meta"]["count"] == number
    assert len(response.get[name]) == number


def test_raw_gql_v4_error() -> None:
    client: WeaviateClient = weaviate.connect_to_local()

    response = client.graphql_raw_query(
        """{
                                        Get{IDoNotExist{name}}
                                        }"""
    )

    assert response.errors is not None
    assert len(response.get) == 0
    assert len(response.aggregate) == 0
    assert len(response.explore) == 0



import datetime
import struct
import time
import uuid as uuid_package
from typing import Any, Dict, List, Optional, Union, cast

import grpc  # type: ignore
from google.protobuf.struct_pb2 import Struct

from weaviate.collections.classes.batch import (
    ErrorObject,
    _BatchObject,
    BatchObjectReturn,
)
from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.collections.classes.types import GeoCoordinate, PhoneNumber
from weaviate.collections.classes.internal import ReferenceToMulti, ReferenceInputs
from weaviate.collections.grpc.shared import _BaseGRPC
from weaviate.connect import ConnectionV4
from weaviate.exceptions import (
    WeaviateBatchError,
    WeaviateInsertInvalidPropertyError,
    WeaviateInsertManyAllFailedError,
    WeaviateInvalidInputError,
)
from weaviate.proto.v1 import batch_pb2, base_pb2
from weaviate.util import _datetime_to_string, _get_vector_v4


def _pack_named_vectors(vectors: Dict[str, List[float]]) -> List[base_pb2.Vectors]:
    return [
        base_pb2.Vectors(
            name=name,
            vector_bytes=struct.pack("{}f".format(len(vector)), *vector),
        )
        for name, vector in vectors.items()
    ]


class _BatchGRPC(_BaseGRPC):
    """This class is used to insert multiple objects into Weaviate using the gRPC API.

    It is used within the `_Data` and `_Batch` classes hence the necessary generalities
    and abstractions so as not to couple to strongly to either use-case.
    """

    def __init__(self, connection: ConnectionV4, consistency_level: Optional[ConsistencyLevel]):
        super().__init__(connection, consistency_level)

    def __grpc_objects(self, objects: List[_BatchObject]) -> List[batch_pb2.BatchObject]:
        def pack_vector(vector: Any) -> bytes:
            vector_list = _get_vector_v4(vector)
            return struct.pack("{}f".format(len(vector_list)), *vector_list)

        return [
            batch_pb2.BatchObject(
                collection=obj.collection,
                vector_bytes=(
                    pack_vector(obj.vector)
                    if obj.vector is not None and isinstance(obj.vector, list)
                    else None
                ),
                uuid=str(obj.uuid) if obj.uuid is not None else str(uuid_package.uuid4()),
                properties=(
                    self.__translate_properties_from_python_to_grpc(
                        obj.properties,
                        obj.references if obj.references is not None else {},
                    )
                    if obj.properties is not None
                    else None
                ),
                tenant=obj.tenant,
                vectors=(
                    _pack_named_vectors(obj.vector)
                    if obj.vector is not None and isinstance(obj.vector, dict)
                    else None
                ),
            )
            for obj in objects
        ]

    def objects(self, objects: List[_BatchObject], timeout: int) -> BatchObjectReturn:
        """Insert multiple objects into Weaviate through the gRPC API.

        Parameters:
            `objects`
                A list of `WeaviateObject` containing the data of the objects to be inserted. The class name must be
                provided for each object, and the UUID is optional. If no UUID is provided, one will be generated for each object.
                The UUIDs of the inserted objects will be returned in the `uuids` attribute of the returned `_BatchReturn` object.
                The UUIDs of the objects that failed to be inserted will be returned in the `errors` attribute of the returned `_BatchReturn` object.
            `tenant`
                The tenant to be used for this batch operation
        """
        weaviate_objs = self.__grpc_objects(objects)

        start = time.time()
        errors = self.__send_batch(weaviate_objs, timeout=timeout)
        elapsed_time = time.time() - start

        if len(errors) == len(weaviate_objs):
            # Escape sequence (backslash) not allowed in expression portion of f-string prior to Python 3.12: pylance
            raise WeaviateInsertManyAllFailedError(
                "Here is the set of all errors: {}".format(
                    "\n".join(err for err in set(errors.values()))
                )
            )

        all_responses: List[Union[uuid_package.UUID, ErrorObject]] = cast(
            List[Union[uuid_package.UUID, ErrorObject]], list(range(len(weaviate_objs)))
        )
        return_success: Dict[int, uuid_package.UUID] = {}
        return_errors: Dict[int, ErrorObject] = {}

        for idx, obj in enumerate(weaviate_objs):
            if idx in errors:
                error = ErrorObject(errors[idx], objects[idx], original_uuid=objects[idx].uuid)
                return_errors[idx] = error
                all_responses[idx] = error
            else:
                success = uuid_package.UUID(obj.uuid)
                return_success[idx] = success
                all_responses[idx] = success

        return BatchObjectReturn(
            uuids=return_success,
            errors=return_errors,
            has_errors=len(errors) > 0,
            all_responses=all_responses,
            elapsed_seconds=elapsed_time,
        )

    def __send_batch(self, batch: List[batch_pb2.BatchObject], timeout: int) -> Dict[int, str]:
        metadata = self._get_metadata()
        try:
            assert self._connection.grpc_stub is not None
            res: batch_pb2.BatchObjectsReply
            res, _ = self._connection.grpc_stub.BatchObjects.with_call(
                batch_pb2.BatchObjectsRequest(
                    objects=batch,
                    consistency_level=self._consistency_level,
                ),
                metadata=metadata,
                timeout=timeout,
            )

            objects: Dict[int, str] = {}
            for result in res.errors:
                objects[result.index] = result.error
            return objects
        except grpc.RpcError as e:
            raise WeaviateBatchError(e.details())  # pyright: ignore

    async def aobjects(self, objects: List[_BatchObject], timeout: int) -> BatchObjectReturn:
        """Insert multiple objects into Weaviate through the gRPC API.

        Parameters:
            `objects`
                A list of `WeaviateObject` containing the data of the objects to be inserted. The class name must be
                provided for each object, and the UUID is optional. If no UUID is provided, one will be generated for each object.
                The UUIDs of the inserted objects will be returned in the `uuids` attribute of the returned `_BatchReturn` object.
                The UUIDs of the objects that failed to be inserted will be returned in the `errors` attribute of the returned `_BatchReturn` object.
            `tenant`
                The tenant to be used for this batch operation
        """
        weaviate_objs = self.__grpc_objects(objects)

        start = time.time()
        errors = await self.__send_batch_async(weaviate_objs, timeout=timeout)
        elapsed_time = time.time() - start

        all_responses: List[Union[uuid_package.UUID, ErrorObject]] = cast(
            List[Union[uuid_package.UUID, ErrorObject]], list(range(len(weaviate_objs)))
        )
        return_success: Dict[int, uuid_package.UUID] = {}
        return_errors: Dict[int, ErrorObject] = {}

        for idx, obj in enumerate(weaviate_objs):
            if idx in errors:
                error = ErrorObject(errors[idx], objects[idx], original_uuid=objects[idx].uuid)
                return_errors[idx] = error
                all_responses[idx] = error
            else:
                success = uuid_package.UUID(obj.uuid)
                return_success[idx] = success
                all_responses[idx] = success

        return BatchObjectReturn(
            uuids=return_success,
            errors=return_errors,
            has_errors=len(errors) > 0,
            all_responses=all_responses,
            elapsed_seconds=elapsed_time,
        )

    async def __send_batch_async(
        self, batch: List[batch_pb2.BatchObject], timeout: int
    ) -> Dict[int, str]:
        metadata = self._get_metadata()
        try:
            assert self._connection.agrpc_stub is not None
            res: batch_pb2.BatchObjectsReply
            res = await self._connection.agrpc_stub.BatchObjects(
                batch_pb2.BatchObjectsRequest(
                    objects=batch,
                    consistency_level=self._consistency_level,
                ),
                metadata=metadata,
                timeout=timeout,
            )

            objects: Dict[int, str] = {}
            for result in res.errors:
                objects[result.index] = result.error
            return objects
        except grpc.RpcError as e:
            raise WeaviateBatchError(e.details())  # pyright: ignore

    def __translate_properties_from_python_to_grpc(
        self, data: Dict[str, Any], refs: ReferenceInputs
    ) -> batch_pb2.BatchObject.Properties:
        _validate_props(data)

        multi_target: List[batch_pb2.BatchObject.MultiTargetRefProps] = []
        single_target: List[batch_pb2.BatchObject.SingleTargetRefProps] = []
        non_ref_properties: Struct = Struct()
        bool_arrays: List[base_pb2.BooleanArrayProperties] = []
        text_arrays: List[base_pb2.TextArrayProperties] = []
        int_arrays: List[base_pb2.IntArrayProperties] = []
        float_arrays: List[base_pb2.NumberArrayProperties] = []
        object_properties: List[base_pb2.ObjectProperties] = []
        object_array_properties: List[base_pb2.ObjectArrayProperties] = []
        empty_lists: List[str] = []

        for key, ref in refs.items():
            if isinstance(ref, ReferenceToMulti):
                multi_target.append(
                    batch_pb2.BatchObject.MultiTargetRefProps(
                        uuids=ref.uuids_str, target_collection=ref.target_collection, prop_name=key
                    )
                )
            elif isinstance(ref, str) or isinstance(ref, uuid_package.UUID):
                single_target.append(
                    batch_pb2.BatchObject.SingleTargetRefProps(uuids=[str(ref)], prop_name=key)
                )
            elif isinstance(ref, list):
                single_target.append(
                    batch_pb2.BatchObject.SingleTargetRefProps(
                        uuids=[str(v) for v in ref], prop_name=key
                    )
                )
            else:
                raise WeaviateInvalidInputError(f"Invalid reference: {ref}")

        for key, entry in data.items():
            if isinstance(entry, dict):
                parsed = self.__translate_properties_from_python_to_grpc(entry, {})
                object_properties.append(
                    base_pb2.ObjectProperties(
                        prop_name=key,
                        value=base_pb2.ObjectPropertiesValue(
                            non_ref_properties=parsed.non_ref_properties,
                            int_array_properties=parsed.int_array_properties,
                            text_array_properties=parsed.text_array_properties,
                            number_array_properties=parsed.number_array_properties,
                            boolean_array_properties=parsed.boolean_array_properties,
                            object_properties=parsed.object_properties,
                            object_array_properties=parsed.object_array_properties,
                            empty_list_props=parsed.empty_list_props,
                        ),
                    )
                )
            elif isinstance(entry, list) and len(entry) == 0:
                empty_lists.append(key)
            elif isinstance(entry, list) and isinstance(entry[0], dict):
                entry = cast(List[Dict[str, Any]], entry)
                object_array_properties.append(
                    base_pb2.ObjectArrayProperties(
                        values=[
                            base_pb2.ObjectPropertiesValue(
                                non_ref_properties=parsed.non_ref_properties,
                                int_array_properties=parsed.int_array_properties,
                                text_array_properties=parsed.text_array_properties,
                                number_array_properties=parsed.number_array_properties,
                                boolean_array_properties=parsed.boolean_array_properties,
                                object_properties=parsed.object_properties,
                                object_array_properties=parsed.object_array_properties,
                                empty_list_props=parsed.empty_list_props,
                            )
                            for v in entry
                            if (parsed := self.__translate_properties_from_python_to_grpc(v, {}))
                        ],
                        prop_name=key,
                    )
                )
            elif isinstance(entry, list) and isinstance(entry[0], bool):
                bool_arrays.append(base_pb2.BooleanArrayProperties(prop_name=key, values=entry))
            elif isinstance(entry, list) and isinstance(entry[0], str):
                text_arrays.append(base_pb2.TextArrayProperties(prop_name=key, values=entry))
            elif isinstance(entry, list) and isinstance(entry[0], datetime.datetime):
                text_arrays.append(
                    base_pb2.TextArrayProperties(
                        prop_name=key, values=[_datetime_to_string(x) for x in entry]
                    )
                )
            elif isinstance(entry, list) and isinstance(entry[0], uuid_package.UUID):
                text_arrays.append(
                    base_pb2.TextArrayProperties(prop_name=key, values=[str(x) for x in entry])
                )
            elif isinstance(entry, list) and isinstance(entry[0], int):
                int_arrays.append(base_pb2.IntArrayProperties(prop_name=key, values=entry))
            elif isinstance(entry, list) and isinstance(entry[0], float):
                values_bytes = struct.pack("{}d".format(len(entry)), *entry)
                float_arrays.append(
                    base_pb2.NumberArrayProperties(prop_name=key, values_bytes=values_bytes)
                )
            elif isinstance(entry, GeoCoordinate):
                non_ref_properties.update({key: entry._to_dict()})
            elif isinstance(entry, PhoneNumber):
                non_ref_properties.update({key: entry._to_dict()})
            else:
                non_ref_properties.update({key: _serialize_primitive(entry)})

        return batch_pb2.BatchObject.Properties(
            non_ref_properties=non_ref_properties,
            multi_target_ref_props=multi_target,
            single_target_ref_props=single_target,
            text_array_properties=text_arrays,
            number_array_properties=float_arrays,
            int_array_properties=int_arrays,
            boolean_array_properties=bool_arrays,
            object_properties=object_properties,
            object_array_properties=object_array_properties,
            empty_list_props=empty_lists,
        )


def _validate_props(props: Dict[str, Any]) -> None:
    if "id" in props or "vector" in props:
        raise WeaviateInsertInvalidPropertyError(props)


def _serialize_primitive(value: Any) -> Any:
    if isinstance(value, uuid_package.UUID):
        return str(value)
    if isinstance(value, datetime.datetime):
        return _datetime_to_string(value)
    if isinstance(value, list):
        return [_serialize_primitive(val) for val in value]

    return value
